# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import numpy as np
from scipy import stats
from pytest import raises
from kwant import rmt
from kwant._common import ensure_rng
assert_allclose = np.testing.assert_allclose


def test_gaussian_symmetries():
    rng = ensure_rng(10)
    for n in (5, 8, 100, 200):
        for sym in rmt.sym_list:
            if sym not in ('A', 'D', 'AI') and n % 2:
                raises(ValueError, rmt.gaussian, 5, sym)
                continue
            h = rmt.gaussian(n, sym, rng=rng)
            if rmt.t(sym):
                t_mat = np.array(rmt.h_t_matrix[sym])
                t_mat = np.kron(np.identity(n // len(t_mat)), t_mat)
                assert_allclose(h, np.dot(t_mat, np.dot(h.conj(), t_mat)),
                                err_msg='TRS broken in ' + sym)
            if rmt.p(sym):
                p_mat = np.array(rmt.h_p_matrix[sym])
                p_mat = np.kron(np.identity(n // len(p_mat)), p_mat)
                assert_allclose(h, -np.dot(p_mat, np.dot(h.conj(), p_mat)),
                                err_msg='PHS broken in ' + sym)
            if rmt.c(sym):
                sz = np.kron(np.identity(n // 2), np.diag([1, -1]))
                assert_allclose(h, -np.dot(sz, np.dot(h, sz)),
                                err_msg='SLS broken in ' + sym)


def test_gaussian_distributions():
    rng = ensure_rng(1)
    n = 8
    for sym in rmt.sym_list:
        matrices = np.array([rmt.gaussian(n, sym, rng=rng)[-1, 0] for i in range(3000)])
        matrices = matrices.imag if sym in ('D', 'BDI') else matrices.real
        ks = stats.kstest(matrices, 'norm')
        assert (ks[1] > 0.1), (sym, ks)


def test_circular():
    rng = ensure_rng(10)
    n = 6
    sy = np.kron(np.identity(n // 2), [[0, 1j], [-1j, 0]])
    for sym in rmt.sym_list:
        if rmt.t(sym) == -1 or rmt.p(sym) == -1:
            raises(ValueError, rmt.circular, 5, sym)
        s = rmt.circular(n, sym, rng=rng)
        assert_allclose(np.dot(s, s.T.conj()), np.identity(n), atol=1e-9,
                        err_msg='Unitarity broken in ' + sym)
        if rmt.t(sym):
            s1 = np.copy(s.T if rmt.p(sym) != -1
                         else np.dot(sy, np.dot(s.T, sy)))
            s1 *= rmt.t(sym) * (-1 if rmt.p(sym) == -1 else 1)
            assert_allclose(s, s1, atol=1e-9, err_msg='TRS broken in ' + sym)
        if rmt.p(sym):
            s1 = np.copy(s.conj() if rmt.p(sym) != -1
                         else np.dot(sy, np.dot(s.conj(), sy)))
            if sym in ('DIII', 'CI'):
                s1 *= -1
            assert_allclose(s, s1, atol=1e-9, err_msg='PHS broken in ' + sym)
        if rmt.c(sym):
            assert_allclose(s, s.T.conj(), atol=1e-9,
                            err_msg='SLS broken in ' + sym)

    # Check for distribution properties if the ensemble is a symmetric group.
    f = lambda x: x[0] / np.linalg.norm(x)
    for sym in ('A', 'C'):
        sample_distr = np.apply_along_axis(f, 0, rng.randn(2 * n, 1000))
        s_sample = np.array([rmt.circular(n, sym, rng=rng)
                             for i in range(1000)])
        assert stats.ks_2samp(sample_distr, s_sample[:, 0, 0].real)[1] > 0.1, \
                'Noncircular distribution in ' + sym
        assert stats.ks_2samp(sample_distr, s_sample[:, 3, 2].real)[1] > 0.1, \
                'Noncircular distribution in ' + sym
        assert stats.ks_2samp(sample_distr, s_sample[:, 1, 1].imag)[1] > 0.1, \
                'Noncircular distribution in ' + sym
        assert stats.ks_2samp(sample_distr, s_sample[:, 2, 3].imag)[1] > 0.1, \
                'Noncircular distribution in ' + sym

    sample_distr = np.apply_along_axis(f, 0, rng.randn(n, 500))
    s_sample = np.array([rmt.circular(n, 'D', rng=rng) for i in range(500)])
    ks = stats.ks_2samp(sample_distr, s_sample[:, 0, 0])
    assert ks[1] > 0.1, 'Noncircular distribution in D ' + str(ks)
    ks = stats.ks_2samp(sample_distr, s_sample[:, 3, 2])
    assert ks[1] > 0.1, 'Noncircular distribution in D ' + str(ks)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import tempfile
import warnings
import itertools
import numpy as np
import tinyarray as ta
from math import cos, sin
import scipy.integrate
import scipy.stats
import pytest
import sys

import kwant
from kwant._common import ensure_rng

try:
    from mpl_toolkits import mplot3d
    import matplotlib

    # This check is the same as the one performed inside matplotlib.use.
    matplotlib_backend_chosen = 'matplotlib.backends' in sys.modules
    # If the user did not already choose a backend, then choose
    # the one with the least dependencies.
    if not matplotlib_backend_chosen:
        matplotlib.use('Agg')

    from matplotlib import pyplot  # pragma: no flakes
except ImportError:
    matplotlib_backend_chosen = False

from kwant import plotter
from kwant import _plotter  # for mpl_available


def test_matplotlib_backend_unset():
    """Simply importing Kwant should not set the matplotlib backend."""
    assert matplotlib_backend_chosen is False


def test_importable_without_matplotlib():
    prefix, sep, suffix = _plotter.__file__.rpartition('.')
    if suffix in ['pyc', 'pyo']:
        suffix = 'py'
    assert suffix == 'py'
    fname = sep.join((prefix, suffix))
    with open(fname, 'rb') as f:
        code = f.read()
    code = code.replace(b'from . import', b'from kwant import')
    code = code.replace(b'matplotlib', b'totalblimp')

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        exec(code)               # Trigger the warning.
        assert len(w) == 1
        assert issubclass(w[0].category, RuntimeWarning)
        assert "only iterator-providing functions" in str(w[0].message)


def syst_2d(W=3, r1=3, r2=8):
    a = 1
    t = 1.0
    lat = kwant.lattice.square(a, norbs=1)
    syst = kwant.Builder()

    def ring(pos):
        (x, y) = pos
        rsq = x ** 2 + y ** 2
        return r1 ** 2 < rsq < r2 ** 2

    syst[lat.shape(ring, (0, r1 + 1))] = 4 * t
    syst[lat.neighbors()] = -t
    sym_lead0 = kwant.TranslationalSymmetry(lat.vec((-1, 0)))
    lead0 = kwant.Builder(sym_lead0)

    lead_shape = lambda pos: (-W / 2 < pos[1] < W / 2)

    lead0[lat.shape(lead_shape, (0, 0))] = 4 * t
    lead0[lat.neighbors()] = - t
    lead1 = lead0.reversed()
    syst.attach_lead(lead0)
    syst.attach_lead(lead1)
    return syst


def syst_3d(W=3, r1=2, r2=4, a=1, t=1.0):
    lat = kwant.lattice.general(((a, 0, 0), (0, a, 0), (0, 0, a)))
    syst = kwant.Builder()

    def ring(pos):
        (x, y, z) = pos
        rsq = x ** 2 + y ** 2
        return (r1 ** 2 < rsq < r2 ** 2) and abs(z) < 2
    syst[lat.shape(ring, (0, -r2 + 1, 0))] = 4 * t
    syst[lat.neighbors()] = - t
    sym_lead0 = kwant.TranslationalSymmetry(lat.vec((-1, 0, 0)))
    lead0 = kwant.Builder(sym_lead0)

    lead_shape = lambda pos: (-W / 2 < pos[1] < W / 2) and abs(pos[2]) < 2

    lead0[lat.shape(lead_shape, (0, 0, 0))] = 4 * t
    lead0[lat.neighbors()] = - t
    lead1 = lead0.reversed()
    syst.attach_lead(lead0)
    syst.attach_lead(lead1)
    return syst


@pytest.mark.skipif(not _plotter.mpl_available, reason="Matplotlib unavailable.")
def test_plot():
    plot = plotter.plot
    syst2d = syst_2d()
    syst3d = syst_3d()
    color_opts = ['k', (lambda site: site.tag[0]),
                  lambda site: (abs(site.tag[0] / 100),
                                abs(site.tag[1] / 100), 0)]
    with tempfile.TemporaryFile('w+b') as out:
        for color in color_opts:
            for syst in (syst2d, syst3d):
                fig = plot(syst, site_color=color, cmap='binary', file=out)
                if (color != 'k' and
                    isinstance(color(next(iter(syst2d.sites()))), float)):
                    assert fig.axes[0].collections[0].get_array() is not None
                assert len(fig.axes[0].collections) == 6
        color_opts = ['k', (lambda site, site2: site.tag[0]),
                      lambda site, site2: (abs(site.tag[0] / 100),
                                           abs(site.tag[1] / 100), 0)]
        for color in color_opts:
            for syst in (syst2d, syst3d):
                fig = plot(syst2d, hop_color=color, cmap='binary', file=out,
                           fig_size=(2, 10), dpi=30)
                if color != 'k' and isinstance(color(next(iter(syst2d.sites())),
                                                          None), float):
                    assert fig.axes[0].collections[1].get_array() is not None

        assert isinstance(plot(syst3d, file=out).axes[0], mplot3d.axes3d.Axes3D)

        syst2d.leads = []
        plot(syst2d, file=out)
        del syst2d[list(syst2d.hoppings())]
        plot(syst2d, file=out)

        plot(syst3d, file=out)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            plot(syst2d.finalized(), file=out)

        # test 2D projections of 3D systems
        plot(syst3d, file=out, pos_transform=lambda pos: pos[:2])

def good_transform(pos):
    x, y = pos
    return y, x

def bad_transform(pos):
    x, y = pos
    return x, y, 0

@pytest.mark.skipif(not _plotter.mpl_available, reason="Matplotlib unavailable.")
def test_map():
    syst = syst_2d()
    with tempfile.TemporaryFile('w+b') as out:
        plotter.map(syst, lambda site: site.tag[0], pos_transform=good_transform,
                    file=out, method='linear', a=4, oversampling=4, cmap='flag')
        pytest.raises(ValueError, plotter.map, syst,
                      lambda site: site.tag[0],
                      pos_transform=bad_transform, file=out)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            plotter.map(syst.finalized(), range(len(syst.sites())),
                        file=out)
        pytest.raises(ValueError, plotter.map, syst,
                      range(len(syst.sites())), file=out)


def test_mask_interpolate():
    # A coordinate array with coordinates of two points almost coinciding.
    coords = np.array([[0, 0], [1e-7, 1e-7], [1, 1], [1, 0]])

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")
        plotter.mask_interpolate(coords, np.ones(len(coords)), a=1)
        assert len(w) == 1
        assert issubclass(w[-1].category, RuntimeWarning)
        assert "coinciding" in str(w[-1].message)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        pytest.raises(ValueError, plotter.mask_interpolate,
                      coords, np.ones(len(coords)))
        pytest.raises(ValueError, plotter.mask_interpolate,
                      coords, np.ones(2 * len(coords)))


@pytest.mark.skipif(not _plotter.mpl_available, reason="Matplotlib unavailable.")
def test_bands():

    syst = syst_2d().finalized().leads[0]

    with tempfile.TemporaryFile('w+b') as out:
        plotter.bands(syst, file=out)
        plotter.bands(syst, fig_size=(10, 10), file=out)
        plotter.bands(syst, momenta=np.linspace(0, 2 * np.pi), file=out)

        fig = pyplot.Figure()
        ax = fig.add_subplot(1, 1, 1)
        plotter.bands(syst, ax=ax, file=out)


@pytest.mark.skipif(not _plotter.mpl_available, reason="Matplotlib unavailable.")
def test_spectrum():

    def ham_1d(a, b, c):
        return a**2 + b**2 + c**2

    def ham_2d(a, b, c):
        return np.eye(2) * (a**2 + b**2 + c**2)

    lat = kwant.lattice.chain()
    syst = kwant.Builder()
    syst[(lat(i) for i in range(3))] = lambda site, a, b: a + b
    syst[lat.neighbors()] = lambda site1, site2, c: c
    fsyst = syst.finalized()

    vals = np.linspace(0, 1, 3)

    with tempfile.TemporaryFile('w+b') as out:

        for ham in (ham_1d, ham_2d, fsyst):
            plotter.spectrum(ham, ('a', vals), params=dict(b=1, c=1), file=out)
            # test with explicit figsize
            plotter.spectrum(ham, ('a', vals), params=dict(b=1, c=1),
                             fig_size=(10, 10), file=out)

        for ham in (ham_1d, ham_2d, fsyst):
            plotter.spectrum(ham, ('a', vals), ('b', 2 * vals),
                             params=dict(c=1), file=out)
            # test with explicit figsize
            plotter.spectrum(ham, ('a', vals), ('b', 2 * vals),
                             params=dict(c=1), fig_size=(10, 10), file=out)

        # test 2D plot and explicitly passing axis
        fig = pyplot.Figure()
        ax = fig.add_subplot(1, 1, 1, projection='3d')
        plotter.spectrum(ham_1d, ('a', vals), ('b', 2 * vals),
                         params=dict(c=1), ax=ax, file=out)
        # explicitly pass axis without 3D support
        ax = fig.add_subplot(1, 1, 1)
        with pytest.raises(TypeError):
            plotter.spectrum(ham_1d, ('a', vals), ('b', 2 * vals),
                             params=dict(c=1), ax=ax, file=out)

    def mask(a, b):
        return a > 0.5

    with tempfile.TemporaryFile('w+b') as out:
        plotter.spectrum(ham, ('a', vals), ('b', 2 * vals), params=dict(c=1),
                         mask=mask, file=out)


def syst_rect(lat, salt, W=3, L=50):
    syst = kwant.Builder()

    ll = L//2
    ww = W//2

    def onsite(site):
        return 4 + 0.1 * kwant.digest.gauss(repr(site.tag), salt=salt)

    syst[(lat(i, j) for i in range(-ll, ll+1)
         for j in range(-ww, ww+1))] = onsite
    syst[lat.neighbors()] = -1

    sym = kwant.TranslationalSymmetry(lat.vec((-1, 0)))
    lead = kwant.Builder(sym)
    lead[(lat(0, j) for j in range(-ww, ww + 1))] = 4
    lead[lat.neighbors()] = -1

    syst.attach_lead(lead)
    syst.attach_lead(lead.reversed())

    return syst


def div(F, h):
    """Calculate the divergence of a vector field F over a grid of spacing h."""
    assert len(F.shape[:-1]) == F.shape[-1]
    assert len(h) == F.shape[-1]
    return sum(np.gradient(F[..., i], h[i])[i] for i in range(F.shape[-1]))


def rotational_currents(g):
    """Return a basis of divergence-free currents for a closed graph.

    Given the graph 'g' of a Kwant system, returns a sequence of arrays
    which are linearly independent, divergence-free currents on the graph.
    """
    #'A' represents the set of expressions that give the net current flow
    # into the system sites. 'perm' is a map from the edges of a graph
    # with only 1 edge per hopping to the proper Kwant graph (2 edges
    # per hopping).
    A = np.zeros((g.num_nodes, g.num_edges // 2))
    hoppings = dict()
    perm_data = np.zeros(g.num_edges, dtype=int)
    perm_ij = np.zeros((2, g.num_edges), dtype=int)
    i = 0
    for k, (a, b) in enumerate(g):
        hop = frozenset((a, b))
        if hop not in hoppings:
            A[a, i] = 1
            A[b, i] = -1
            hoppings[hop] = i
            perm_data[k] = 1
            perm_ij[:, k] = (k, i)
            i += 1
        else:
            perm_data[k] = -1
            perm_ij[:, k] = (k, hoppings[hop])

    perm = scipy.sparse.coo_matrix((perm_data, perm_ij))

    # Get the row vectors of V with singular value 0. These form
    # a basis for the right null space of 'A'.
    U, S, V = np.linalg.svd(A)
    tol = S.max() * max(A.shape) * np.finfo(S.dtype).eps
    rank = sum(S > tol)
    # Transform null space basis into vectors defined over the full
    # hopping space (both hopping directions).
    null_space_basis = V[-(len(V) - rank):].transpose()
    null_space_basis = perm.dot(null_space_basis).transpose()
    return null_space_basis


def _border_is_0(field):
    borders = [(0, slice(None)), (-1, slice(None)),
               (slice(None), 0), (slice(None), -1)]
    return all(np.allclose(field[a, b], 0) for a, b in borders)


def _test_border_0(interpolator):
    ## Test that current is always identically zero at box boundaries
    syst = kwant.Builder()
    lat = kwant.lattice.square()
    syst[[lat(0, 0), lat(1, 0)]] = None
    syst[(lat(0, 0), lat(1, 0))] = None
    syst = syst.finalized()
    values = [1, -1]

    ns = [3, 4, 5, 10, 100]
    abswidths = [0.01, 0.1, 1, 10, 100]
    relwidths = [0.01, 0.1, 1, 10, 100]
    for n, abswidth in itertools.product(ns, abswidths):
        field, _ = interpolator(syst, values, abswidth=abswidth, n=n)
        assert _border_is_0(field)
    for n, relwidth in itertools.product(ns, relwidths):
        field, _ = interpolator(syst, values, relwidth=relwidth, n=n)
        assert _border_is_0(field)


def test_density_interpolation():
    ## Passing a Builder will raise an error
    pytest.raises(TypeError, plotter.interpolate_density, syst_2d(), None)

    # Test that the density is always identically zero at the box boundaries
    # as the bump function has finite support and we add a padding
    _test_border_0(kwant.plotter.interpolate_density)

    def R(theta):
        return ta.array([[cos(theta), -sin(theta)], [sin(theta), cos(theta)]])

    # Make lattice with lattice vectors perturbed from x and y directions
    def make_lattice(a, salt='0'):
        theta_x = kwant.digest.uniform('x', salt=salt) * np.pi / 6
        theta_y = kwant.digest.uniform('y', salt=salt) * np.pi / 6
        x = ta.dot(R(theta_x), (a, 0))
        y = ta.dot(R(theta_y), (0, a))
        return kwant.lattice.general([x, y], norbs=1)

    # Check that integrating the interpolated density gives the same result
    # as summing the densities on all sites. We check this for several lattice
    # widths, lattice orientations and bump widths.
    for a, width in itertools.product((1, 2), (1, 0.5)):
        lat = make_lattice(a)
        syst = syst_rect(lat, salt='0').finalized()

        psi = kwant.wave_function(syst, energy=3)(0)[0]
        density = kwant.operator.Density(syst)(psi)
        exact_charge = sum(density)
        # We verify that the result is good by interpolating for
        # various numbers of points-per-bump and verifying that
        # the error falls of as 1/n.
        data = []
        for n in [4, 6, 8, 11, 16]:
            rho, box = plotter.interpolate_density(syst, density,
                                                   n=n, abswidth=width)
            (xmin, xmax), (ymin, ymax) = box
            area =  xmax - xmin * (ymax - ymin)
            N = rho.shape[0] * rho.shape[1]
            charge = np.sum(rho) * area / N
            data.append((n, abs(charge - exact_charge)))
        _, _, rvalue, *_ = scipy.stats.linregress(np.log(data))
        # Gradient of -1 on log-log plot means error falls off as 1/n
        assert rvalue < -0.8

    # Test that the interpolation is linear in the input.
    rng = ensure_rng(1)
    lat = make_lattice(1, '1')
    syst = syst_rect(lat, salt='1').finalized()
    rho_0 = rng.rand(len(syst.sites))
    rho_1 = rng.rand(len(syst.sites))

    irho_0, _ = plotter.interpolate_density(syst, rho_0)
    irho_1, _ = plotter.interpolate_density(syst, rho_1)

    rho_tot, _ = plotter.interpolate_density(syst, rho_0 + 2 * rho_1)
    assert np.allclose(rho_tot, irho_0 + 2 * irho_1)


def test_current_interpolation():

    ## Passing a Builder will raise an error
    pytest.raises(TypeError, plotter.interpolate_current, syst_2d(), None)

    def R(theta):
        return ta.array([[cos(theta), -sin(theta)], [sin(theta), cos(theta)]])

    def make_lattice(a, theta):
        x = ta.dot(R(theta), (a, 0))
        y = ta.dot(R(theta), (0, a))
        return kwant.lattice.general([x, y], norbs=1)

    _test_border_0(plotter.interpolate_current)

    ## Check current through cross section is same for different lattice
    ## parameters and orientations of the system wrt. the discretization grid
    for a, theta, width in [(1, 0, 1),
                            (1, 0, 0.5),
                            (2, 0, 1),
                            (1, 0.2, 1),
                            (2, 0.4, 1)]:
        lat = make_lattice(a, theta)
        syst = syst_rect(lat, salt='0').finalized()
        psi = kwant.wave_function(syst, energy=3)(0)

        def cut(a, b):
            return b.tag[0] < 0 and a.tag[0] >= 0

        J = kwant.operator.Current(syst).bind()
        J_cut = kwant.operator.Current(syst, where=cut, sum=True).bind()
        J_exact = J_cut(psi[0])

        data = []
        for n in [4, 6, 8, 11, 16]:
            j0, box = plotter.interpolate_current(syst, J(psi[0]),
                                                  n=n, abswidth=width)
            x, y = (np.linspace(mn, mx, shape)
                    for (mn, mx), shape in zip(box, j0.shape))
            # slice field perpendicular to a cut along the y axis
            y_axis = (np.argmin(np.abs(x)), slice(None), 0)
            J_interp = scipy.integrate.simps(j0[y_axis], y)
            data.append((n, abs(J_interp - J_exact)))
        # 3rd value returned from 'linregress' is 'rvalue'
        assert scipy.stats.linregress(np.log(data))[2] < -0.8


    ### Tests on a divergence-free current (closed system)

    lat = kwant.lattice.general([(1, 0), (0.5, np.sqrt(3) / 2)])
    syst = kwant.Builder()
    sites = [lat(0, 0), lat(1, 0), lat(0, 1), lat(2, 2)]
    syst[sites] = None
    syst[((s, t) for s, t in itertools.product(sites, sites) if s != t)] = None
    del syst[lat(0, 0), lat(2, 2)]
    syst = syst.finalized()

    # generate random divergence-free currents
    Js = rotational_currents(syst.graph)
    rng = ensure_rng(3)
    J0 = sum(rng.rand(len(Js))[:, None] * Js)
    J1 = sum(rng.rand(len(Js))[:, None] * Js)

    # Sanity check that diverence on the graph is 0
    divergence = np.zeros(len(syst.sites))
    for (a, _), current in zip(syst.graph, J0):
        divergence[a] += current
    assert np.allclose(divergence, 0)

    j0, _ = plotter.interpolate_current(syst, J0)
    j1, _ = plotter.interpolate_current(syst, J1)

    ## Test linearity of interpolation.
    j_tot, _ = plotter.interpolate_current(syst, J0 + 2 * J1)
    assert np.allclose(j_tot, j0 + 2 * j1)

    ## Test that divergence of interpolated current approaches zero as we make
    ## the interpolation finer.
    data = []
    for n in [4, 6, 8, 11, 16]:
        j, box = plotter.interpolate_current(syst, J0, n=n)
        dx = [(mx - mn) / (shape - 1) for (mn, mx), shape in zip(box, j.shape)]
        div_j = np.max(np.abs(div(j, dx)))
        data.append((n, div_j))

    # 3rd value returned from 'linregress' is 'rvalue'
    assert scipy.stats.linregress(np.log(data))[2] < -0.8


@pytest.mark.skipif(not _plotter.mpl_available, reason="Matplotlib unavailable.")
def test_current():
    syst = syst_2d().finalized()
    J = kwant.operator.Current(syst)
    current = J(kwant.wave_function(syst, energy=1)(1)[0])

    # Test good codepath
    with tempfile.TemporaryFile('w+b') as out:
        plotter.current(syst, current, file=out)

        fig = pyplot.Figure()
        ax = fig.add_subplot(1, 1, 1)
        plotter.current(syst, current, ax=ax, file=out)



# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['DiscreteSymmetry']

import numpy as np
from scipy.sparse import identity, csr_matrix, hstack


def almost_identity(mat):
    return np.all(abs(mat - identity(mat.shape[0])).data < 1e-10)


def cond_conj(op, conj):
    return op.conj() if conj else op


_conj = [True, True, False]
_names = ['Time reversal', 'Particle-hole', 'Chiral']
_signs = [1, -1, -1]

class DiscreteSymmetry:
    """A collection of discrete symmetries and conservation laws.

    Parameters
    ----------
    projectors : iterable of rectangular or square sparse matrices
        Projectors that block-diagonalize the Hamiltonian.
    time_reversal : square sparse matrix
        The unitary part of the time-reversal symmetry operator.
    particle_hole : square sparse matrix
        The unitary part of the particle-hole symmetry operator.
    chiral : square sparse matrix
        The chiral symmetry operator.

    Notes
    -----
    Whenever one or more discrete symmetry is declared in conjunction with a
    conservation law, the symmetry operators and projectors must be declared
    in canonical form. This means that each block of the Hamiltonian is
    transformed either to itself by a discrete symmetry or to a single
    other block.

    More formally, consider a discrete symmetry S. The symmetry projection
    that maps from block i to block j of the Hamiltonian with projectors
    :math:`P_i` and :math:`P_j` is :math:`S_{ji} = P_j^+ S P_i`.
    If :math:`S_{ji}` is nonzero, a symmetry relation exists between
    blocks i and j. Canonical form means that for each j, the block
    :math:`S_{ji}` is nonzero at most for one i, while all other blocks vanish.

    If the operators are not in canonical form, they can be made so by
    further splitting the Hamiltonian into smaller blocks, i.e. by adding
    more projectors.
    """
    def __init__(self, projectors=None, time_reversal=None, particle_hole=None,
                 chiral=None):
        # Normalize format
        if projectors is not None:
            try:
                projectors = [proj.tocsr() for proj in projectors]
            except AttributeError:
                raise TypeError("projectors must be a sequence of "
                                "sparse matrices.")

        symms = (time_reversal, particle_hole, chiral)
        try:
            symms = [symm.tocsr() if symm is not None else None
                     for symm in symms]
        except AttributeError:
            raise TypeError("Symmetries must be sparse matrices.")

        if None not in (time_reversal, particle_hole, chiral):
            product = chiral.dot(particle_hole).dot(time_reversal.conj())
            if not almost_identity(product):
                raise ValueError("Product of all symmetries "
                                 "should be equal to identity.")

        # Auto-compute a missing symmetry.
        if None not in (time_reversal, particle_hole) and chiral is None:
            chiral = particle_hole.dot(time_reversal.conj())
        elif None not in (particle_hole, chiral) and time_reversal is None:
            # Product of particle hole with itself to get the sign right.
            time_reversal = (particle_hole.dot(particle_hole.conj()).dot(
                             particle_hole).dot(chiral.conj()))
        elif None not in (chiral, time_reversal) and particle_hole is None:
            particle_hole = (time_reversal.dot(time_reversal.conj()).dot(
                             chiral).dot(time_reversal))

        symms = [time_reversal, particle_hole, chiral]
        for i, symm, conj, name in zip(range(3), symms, _conj, _names):
            if symm is None:
                continue
            if not almost_identity(symm.T.conj().dot(symm)):
                raise ValueError("{} symmetry should be "
                                 "{}unitary.".format(name, "anti" * conj))
            symm_squared = symm.dot(cond_conj(symm, conj))
            if not (almost_identity(symm_squared) or
                    (conj and almost_identity(-symm_squared))):
                raise ValueError("{} symmetry should square to "
                                 "{}1.".format(name, "±" * conj))
            symms[i] = symm

        if projectors is not None:
            projectors = [p[:, np.asarray(abs(p).sum(0)).flatten() > 1e-10]
                          for p in projectors]
            if not almost_identity(sum(projector.dot(projector.conj().T) for
                                       projector in projectors)):
                raise ValueError("Projectors should stack to a unitary matrix.")

            # Check that the symmetries and conservation law are in canonical
            # form.  In the projector basis, each discrete symmetry should have
            # at most one nonzero block per row.
            for (symm, name, conj) in zip(symms, _names, _conj):
                if symm is None:
                    continue
                for p_i in projectors:
                    p_hc = p_i.T.conj()
                    nonzero_blocks = sum(
                            np.any(abs(p_hc.dot(symm).dot(
                                (cond_conj(p_j, conj))).data > 1e-7))
                                         for p_j in projectors)
                    if nonzero_blocks > 1:
                        raise ValueError(name + ' symmetry is not in canonical '
                                         'form: see DiscreteSymmetry '
                                         'documentation.')
        self.projectors = projectors
        self.time_reversal, self.particle_hole, self.chiral = symms

    def validate(self, matrix):
        """Check if a matrix satisfies the discrete symmetries.

        Parameters
        ----------
        matrix : sparse or dense matrix
            If rectangular, it is padded by zeros to be square.

        Returns
        -------
        broken_symmetry : string or ``None``
            One of "Conservation law", "Time reversal", "Particle-hole",
            "Chiral": the symmetry broken by the matrix. If the matrix breaks
            more than one symmetry, returns only the first failed check.
        """
        # Extra transposes are to enforse sparse dot product in case matrix is
        # dense.
        n, m = matrix.shape
        if n != m:
            if isinstance(matrix, np.ndarray):
                new_matrix = np.empty((n, n), dtype=matrix.dtype)
                new_matrix[:, :m] = matrix
                new_matrix[:, m:] = 0
                matrix = new_matrix
            else:
                matrix = hstack([matrix, csr_matrix((n, n-m))])
        if self.projectors is not None:
            for proj in self.projectors:
                full = proj.dot(proj.T.conj())
                commutator = full.dot(matrix) - (full.T.dot(matrix.T)).T
                if np.linalg.norm(commutator.data) > 1e-8:
                    return 'Conservation law'
        for symm, conj, sign, name in zip(self[1:], _conj, _signs, _names):
            if symm is None:
                continue
            commutator = symm.T.conj().dot((symm.T.dot(matrix.T)).T)
            commutator = commutator - sign * cond_conj(matrix, conj)
            if np.linalg.norm(commutator.data) > 1e-8:
                return name

    def __getitem__(self, item):
        return (self.projectors, self.time_reversal,
                self.particle_hole, self.chiral)[item]



# Copyright 2011-2014 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['SparseSolver', 'SMatrix', 'GreensFunction']

from collections import namedtuple
from itertools import product
import abc
from  numbers import Integral
import numpy as np
import scipy.sparse as sp
from .._common import ensure_isinstance
from .. import system
from functools import reduce

# Until v0.13.0, scipy.sparse did not support making block matrices out of
# matrices with one dimension being zero:
# https://github.com/scipy/scipy/issues/2127 Additionally, 'scipy.sparse.bmat'
# didn't support matrices with zero size until v0.18:
# https://github.com/scipy/scipy/issues/5976. For now we use NumPy dense
# matrices as a replacement.

# TODO: Once we depend on scipy >= 0.18, code for the special cases can be
# removed from _make_linear_sys, _solve_linear_sys and possibly other places
# marked by the line "See comment about zero-shaped sparse matrices at the top
# of common.py".

LinearSys = namedtuple('LinearSys', ['lhs', 'rhs', 'indices', 'num_orb'])


class SparseSolver(metaclass=abc.ABCMeta):
    """Solver class for computing physical quantities based on solving
    a liner system of equations.

    `SparseSolver` is an abstract base class.  It cannot be instantiated as it
    does not specify the actual linear solve step.  In order to be directly
    usable, a derived class must implement the methods

    - `_factorize` and `_solve_linear_sys`, that solve a linear system of
      equations

    and the following properties:

    - `lhsformat`, `rhsformat`: Sparse matrix format of left and right hand
       sides of the linear system, respectively.  Must be one of {'coo', 'csc',
       'csr'}.

    - `nrhs`: Number of right hand side vectors that should be solved at once
      in a call to `solve_linear_sys`, when the full solution is computed (i.e.
      kept_vars covers all entries in the solution). This should be not too big
      too avoid excessive memory usage, but for some solvers not too small for
      performance reasons.
    """

    @abc.abstractmethod
    def _factorized(self, a):
        """
        Return a preprocessed version of a matrix for the use with
        `solve_linear_sys`.

        Parameters
        ----------
        a : a scipy.sparse.coo_matrix sparse matrix.

        Returns
        -------
        factorized_a : object
            factorized lhs to be used with `_solve_linear_sys`.
        """
        pass

    @abc.abstractmethod
    def _solve_linear_sys(self, factorized_a, b, kept_vars):
        """
        Solve the linar system `a x = b`, returning the part of
        the result indicated in kept_vars.

        Parameters
        ----------
        factorized : object
            The result of calling `_factorized` for the matrix a.
        b : sparse matrix
            The right hand side. Its format must match `rhsformat`.
        kept_vars : slice object or sequence of integers
            A sequence of numbers of variables to keep in the solution

        Returns
        -------
        output : NumPy matrix
            Solution to the system of equations.
        """
        pass

    def _make_linear_sys(self, sys, in_leads, energy=0, args=(),
                         check_hermiticity=True, realspace=False,
                         *, params=None):
        """Make a sparse linear system of equations defining a scattering
        problem.

        Parameters
        ----------
        sys : kwant.system.FiniteSystem
            Low level system, containing the leads and the Hamiltonian of a
            scattering region.
        in_leads : sequence of integers
            Numbers of leads in which current or wave function is injected.
        energy : number
            Excitation energy at which to solve the scattering problem.
        args : tuple, defaults to empty
            Positional arguments to pass to the ``hamiltonian`` method.
            Mutually exclusive with 'params'.
        check_hermiticity : bool
            Check if Hamiltonian matrices are in fact Hermitian.
            Enables deduction of missing transmission coefficients.
        realspace : bool
            Calculate Green's function between the outermost lead
            sites, instead of lead modes. This is almost always
            more computationally expensive and less stable.
        params : dict, optional
            Dictionary of parameter names and their values. Mutually exclusive
            with 'args'.

        Returns
        -------
        (lhs, rhs, indices, num_orb) : LinearSys
            `lhs` is a scipy.sparse.csc_matrix, containing the left hand side
            of the system of equations.  `rhs` is a list of matrices with the
            right hand side, with each matrix corresponding to one lead
            mentioned in `in_leads`. `indices` is a list of arrays of variables
            in the system of equations corresponding to the the outgoing modes
            in each lead, or the indices of variables, on which a lead defined
            via self-energy adds the self-energy. Finally `num_orb` is the
            total number of degrees of freedom in the scattering region.

        lead_info : list of objects
            Contains one entry for each lead.  If `realspace=False`, this is an
            instance of `~kwant.physics.PropagatingModes` with a corresponding
            format, otherwise the lead self-energy matrix.

        Notes
        -----
        All the leads should implement a method `modes` if `realspace=False`
        and a method `selfenergy`.

        The system of equations that is created will be described in detail
        elsewhere.
        """

        syst = sys  # ensure consistent naming across function bodies
        ensure_isinstance(syst, system.System)

        splhsmat = getattr(sp, self.lhsformat + '_matrix')
        sprhsmat = getattr(sp, self.rhsformat + '_matrix')

        if not syst.lead_interfaces:
            raise ValueError('System contains no leads.')
        lhs, norb = syst.hamiltonian_submatrix(args, sparse=True,
                                               return_norb=True,
                                               params=params)[:2]
        lhs = getattr(lhs, 'to' + self.lhsformat)()
        lhs = lhs - energy * sp.identity(lhs.shape[0], format=self.lhsformat)
        num_orb = lhs.shape[0]

        if check_hermiticity and len(lhs.data):
            rtol = 1e-13
            atol = 1e-300
            tol = rtol * np.max(np.abs(lhs.data)) + atol
            if np.any(np.abs((lhs - lhs.T.conj()).data) > tol):
                raise ValueError('System Hamiltonian is not Hermitian. '
                                 'Use option `check_hermiticity=False` '
                                 'if this is intentional.')

        offsets = np.empty(norb.shape[0] + 1, int)
        offsets[0] = 0
        offsets[1 :] = np.cumsum(norb)

        # Process the leads, generate the eigenvector matrices and lambda
        # vectors. Then create blocks of the linear system and add them
        # step by step.
        indices = []
        rhs = []
        lead_info = []
        for leadnum, interface in enumerate(syst.lead_interfaces):
            lead = syst.leads[leadnum]
            if not realspace:
                prop, stab = lead.modes(energy, args, params=params)
                lead_info.append(prop)
                u = stab.vecs
                ulinv = stab.vecslmbdainv
                nprop = stab.nmodes
                svd_v = stab.sqrt_hop

                if len(u) == 0:
                    rhs.append(None)
                    continue

                indices.append(np.arange(lhs.shape[0], lhs.shape[0] + nprop))

                u_out, ulinv_out = u[:, nprop:], ulinv[:, nprop:]
                u_in, ulinv_in = u[:, :nprop], ulinv[:, :nprop]

                # Construct a matrix of 1's that translates the
                # inter-cell hopping to a proper hopping
                # from the system to the lead.
                iface_orbs = np.r_[tuple(slice(offsets[i], offsets[i + 1])
                                        for i in interface)]

                n_lead_orbs = (svd_v.shape[0] if svd_v is not None
                               else u_out.shape[0])
                if n_lead_orbs != len(iface_orbs):
                    msg = ('Lead {0} has hopping with dimensions '
                           'incompatible with its interface dimension.')
                    raise ValueError(msg.format(leadnum))

                coords = np.r_[[np.arange(len(iface_orbs))], [iface_orbs]]
                transf = sp.csc_matrix((np.ones(len(iface_orbs)), coords),
                                       shape=(iface_orbs.size, lhs.shape[0]))

                if svd_v is not None:
                    v_sp = sp.csc_matrix(svd_v.T.conj()) * transf
                    vdaguout_sp = (transf.T *
                                   sp.csc_matrix(np.dot(svd_v, u_out)))
                    lead_mat = - ulinv_out
                else:
                    v_sp = transf
                    vdaguout_sp = transf.T * sp.csc_matrix(u_out)
                    lead_mat = - ulinv_out

                lhs = sp.bmat([[lhs, vdaguout_sp], [v_sp, lead_mat]],
                              format=self.lhsformat)

                if leadnum in in_leads and nprop > 0:
                    if svd_v is not None:
                        vdaguin_sp = transf.T * sp.csc_matrix(
                            -np.dot(svd_v, u_in))
                    else:
                        vdaguin_sp = transf.T * sp.csc_matrix(-u_in)

                    # defer formation of the real matrix until the proper
                    # system size is known
                    rhs.append((vdaguin_sp, ulinv_in))
                else:
                    rhs.append(None)
            else:
                sigma = np.asarray(lead.selfenergy(energy, args, params=params))
                lead_info.append(sigma)
                coords = np.r_[tuple(slice(offsets[i], offsets[i + 1])
                                      for i in interface)]

                if sigma.shape != 2 * coords.shape:
                    msg = ('Self-energy dimension for lead {0} does not '
                           'match the total number of orbitals of the '
                           'sites for which it is defined.')
                    raise ValueError(msg.format(leadnum))

                y, x = np.meshgrid(coords, coords)
                sig_sparse = splhsmat((sigma.flat, [x.flat, y.flat]),
                                      lhs.shape)
                lhs = lhs + sig_sparse  # __iadd__ is not implemented in v0.7
                indices.append(coords)
                if leadnum in in_leads:
                    # defer formation of true rhs until the proper system
                    # size is known
                    rhs.append((coords,))

        # Resize the right-hand sides to be compatible with the full lhs
        for i, mats in enumerate(rhs):
            if isinstance(mats, tuple):
                if len(mats) == 1:
                    # self-energy lead
                    l = mats[0].shape[0]
                    rhs[i] = sprhsmat((-np.ones(l), [mats[0], np.arange(l)]),
                                      shape=(lhs.shape[0], l))
                else:
                    # lead with modes
                    zero_rows = (lhs.shape[0] - mats[0].shape[0] -
                                 mats[1].shape[0])

                    if zero_rows:
                        zero_mat = sprhsmat((zero_rows, mats[0].shape[1]))
                        bmat = [[mats[0]], [mats[1]], [zero_mat]]
                    else:
                        bmat = [[mats[0]], [mats[1]]]

                    rhs[i] = sp.bmat(bmat, format=self.rhsformat)
            elif mats is None:
                # A lead with no rhs.
                rhs[i] = np.zeros((lhs.shape[0], 0))
            else:
                raise RuntimeError('Unknown right-hand side format')

        return LinearSys(lhs, rhs, indices, num_orb), lead_info

    def smatrix(self, sys, energy=0, args=(),
                out_leads=None, in_leads=None, check_hermiticity=True,
                *, params=None):
        """
        Compute the scattering matrix of a system.

        An alias exists for this common name: ``kwant.smatrix``.

        Parameters
        ----------
        sys : `kwant.system.FiniteSystem`
            Low level system, containing the leads and the Hamiltonian of a
            scattering region.
        energy : number
            Excitation energy at which to solve the scattering problem.
        args : tuple, defaults to empty
            Positional arguments to pass to the ``hamiltonian`` method.
            Mutually exclusive with 'params'.
        out_leads : sequence of integers or ``None``
            Numbers of leads where current or wave function is extracted.  None
            is interpreted as all leads. Default is ``None`` and means "all
            leads".
        in_leads : sequence of integers or ``None``
            Numbers of leads in which current or wave function is injected.
            None is interpreted as all leads. Default is ``None`` and means
            "all leads".
        check_hermiticity : ``bool``
            Check if the Hamiltonian matrices are Hermitian.
            Enables deduction of missing transmission coefficients.
        params : dict, optional
            Dictionary of parameter names and their values. Mutually exclusive
            with 'args'.

        Returns
        -------
        output : `~kwant.solvers.common.SMatrix`
            See the notes below and `~kwant.solvers.common.SMatrix`
            documentation.

        Notes
        -----
        This function can be used to calculate the conductance and other
        transport properties of a system.  See the documentation for its output
        type, `~kwant.solvers.common.SMatrix`.

        The returned object contains the scattering matrix elements from the
        `in_leads` to the `out_leads` as well as information about the lead
        modes.

        Both `in_leads` and `out_leads` must be sorted and may only contain
        unique entries.
        """

        syst = sys  # ensure consistent naming across function bodies
        ensure_isinstance(syst, system.System)

        n = len(syst.lead_interfaces)
        if in_leads is None:
            in_leads = list(range(n))
        else:
            in_leads = list(in_leads)
        if out_leads is None:
            out_leads = list(range(n))
        else:
            out_leads = list(out_leads)
        if (np.any(np.diff(in_leads) <= 0) or np.any(np.diff(out_leads) <= 0)):
            raise ValueError("Lead lists must be sorted and "
                             "with unique entries.")
        if len(in_leads) == 0 or len(out_leads) == 0:
            raise ValueError("No output is requested.")

        linsys, lead_info = self._make_linear_sys(syst, in_leads, energy, args,
                                                  check_hermiticity, False,
                                                  params=params)

        kept_vars = np.concatenate([coords for i, coords in
                                    enumerate(linsys.indices) if i in
                                    out_leads])

        # Do not perform factorization if no calculation is to be done.
        len_rhs = sum(i.shape[1] for i in linsys.rhs)
        len_kv = len(kept_vars)
        if not(len_rhs and len_kv):
            return SMatrix(np.zeros((len_kv, len_rhs)), lead_info,
                           out_leads, in_leads, check_hermiticity)

        # See comment about zero-shaped sparse matrices at the top of common.py.
        rhs = sp.bmat([[i for i in linsys.rhs if i.shape[1]]],
                      format=self.rhsformat)
        flhs = self._factorized(linsys.lhs)
        data = self._solve_linear_sys(flhs, rhs, kept_vars)

        return SMatrix(data, lead_info, out_leads, in_leads, check_hermiticity)

    def greens_function(self, sys, energy=0, args=(),
                        out_leads=None, in_leads=None, check_hermiticity=True,
                        *, params=None):
        """
        Compute the retarded Green's function of the system between its leads.

        An alias exists for this common name: ``kwant.greens_function``.

        Parameters
        ----------
        sys : `kwant.system.FiniteSystem`
            Low level system, containing the leads and the Hamiltonian of a
            scattering region.
        energy : number
            Excitation energy at which to solve the scattering problem.
        args : tuple, defaults to empty
            Positional arguments to pass to the ``hamiltonian`` method.
            Mutually exclusive with 'params'.
        out_leads : sequence of integers or ``None``
            Numbers of leads where current or wave function is extracted.  None
            is interpreted as all leads. Default is ``None`` and means "all
            leads".
        in_leads : sequence of integers or ``None``
            Numbers of leads in which current or wave function is injected.
            None is interpreted as all leads. Default is ``None`` and means
            "all leads".
        check_hermiticity : ``bool``
            Check if the Hamiltonian matrices are Hermitian.
            Enables deduction of missing transmission coefficients.
        params : dict, optional
            Dictionary of parameter names and their values. Mutually exclusive
            with 'args'.

        Returns
        -------
        output : `~kwant.solvers.common.GreensFunction`
            See the notes below and `~kwant.solvers.common.GreensFunction`
            documentation.

        Notes
        -----
        This function can be used to calculate the conductance and other
        transport properties of a system.  It is often slower and less stable
        than the scattering matrix-based calculation executed by
        `~kwant.solvers.common.smatrix`, and is currently provided mostly for testing
        purposes and compatibility with RGF code.

        It returns an object encapsulating the Green's function elements
        between the system sites interfacing the leads in `in_leads` and those
        interfacing the leads in `out_leads`.  The returned object also
        contains a list with self-energies of the leads.

        Both `in_leads` and `out_leads` must be sorted and may only contain
        unique entries.
        """

        syst = sys  # ensure consistent naming across function bodies
        ensure_isinstance(syst, system.System)

        n = len(syst.lead_interfaces)
        if in_leads is None:
            in_leads = list(range(n))
        else:
            in_leads = list(in_leads)
        if out_leads is None:
            out_leads = list(range(n))
        else:
            out_leads = list(out_leads)
        if (np.any(np.diff(in_leads) <= 0) or np.any(np.diff(out_leads) <= 0)):
            raise ValueError("Lead lists must be sorted and "
                             "with unique entries.")
        if len(in_leads) == 0 or len(out_leads) == 0:
            raise ValueError("No output is requested.")

        linsys, lead_info = self._make_linear_sys(syst, in_leads, energy, args,
                                                  check_hermiticity, True,
                                                  params=params)

        kept_vars = np.concatenate([coords for i, coords in
                                    enumerate(linsys.indices) if i in
                                    out_leads])

        # Do not perform factorization if no calculation is to be done.
        len_rhs = sum(i.shape[1] for i in linsys.rhs)
        len_kv = len(kept_vars)
        if not(len_rhs and len_kv):
            return GreensFunction(np.zeros((len_kv, len_rhs)), lead_info,
                                  out_leads, in_leads, check_hermiticity)

        # See comment about zero-shaped sparse matrices at the top of common.py.
        rhs = sp.bmat([[i for i in linsys.rhs if i.shape[1]]],
                      format=self.rhsformat)
        flhs = self._factorized(linsys.lhs)
        data = self._solve_linear_sys(flhs, rhs, kept_vars)

        return GreensFunction(data, lead_info, out_leads, in_leads,
                              check_hermiticity)

    def ldos(self, sys, energy=0, args=(), check_hermiticity=True,
             *, params=None):
        """
        Calculate the local density of states of a system at a given energy.

        An alias exists for this common name: ``kwant.ldos``.

        Parameters
        ----------
        sys : `kwant.system.FiniteSystem`
            Low level system, containing the leads and the Hamiltonian of the
            scattering region.
        energy : number
            Excitation energy at which to solve the scattering problem.
        args : tuple of arguments, or empty tuple
            Positional arguments to pass to the function(s) which
            evaluate the hamiltonian matrix elements.  Mutually exclusive
            with 'params'.
        check_hermiticity : ``bool``
            Check if the Hamiltonian matrices are Hermitian.
        params : dict, optional
            Dictionary of parameter names and their values. Mutually exclusive
            with 'args'.

        Returns
        -------
        ldos : a NumPy array
            Local density of states at each orbital of the system.
        """

        syst = sys  # ensure consistent naming across function bodies
        ensure_isinstance(syst, system.System)
        if not check_hermiticity:
            raise NotImplementedError("ldos for non-Hermitian Hamiltonians "
                                      "is not implemented yet.")

        for lead in syst.leads:
            if not hasattr(lead, 'modes') and hasattr(lead, 'selfenergy'):
                # TODO: fix this
                raise NotImplementedError("ldos for leads with only "
                                          "self-energy is not implemented yet.")

        linsys = self._make_linear_sys(syst, range(len(syst.leads)), energy,
                                       args, check_hermiticity,
                                       params=params)[0]

        ldos = np.zeros(linsys.num_orb, float)

        # Do not perform factorization if no further calculation is needed.
        if not sum(i.shape[1] for i in linsys.rhs):
            return ldos

        factored = self._factorized(linsys.lhs)

        # See comment about zero-shaped sparse matrices at the top of common.py.
        rhs = sp.bmat([[i for i in linsys.rhs if i.shape[1]]],
                      format=self.rhsformat)
        for j in range(0, rhs.shape[1], self.nrhs):
            jend = min(j + self.nrhs, rhs.shape[1])
            psi = self._solve_linear_sys(factored, rhs[:, j:jend],
                                         slice(linsys.num_orb))
            ldos += np.sum(np.square(abs(psi)), axis=1)

        return ldos * (0.5 / np.pi)

    def wave_function(self, sys, energy=0, args=(), check_hermiticity=True,
                      *, params=None):
        """
        Return a callable object for the computation of the wave function
        inside the scattering region.

        An alias exists for this common name: ``kwant.wave_function``.

        Parameters
        ----------
        sys : `kwant.system.FiniteSystem`
            The low level system for which the wave functions are to be
            calculated.
        args : tuple of arguments, or empty tuple
            Positional arguments to pass to the function(s) which
            evaluate the hamiltonian matrix elements. Mutually exclusive
            with 'params'.
        check_hermiticity : ``bool``
            Check if the Hamiltonian matrices are Hermitian.
        params : dict, optional
            Dictionary of parameter names and their values. Mutually exclusive
            with 'args'.

        Notes
        -----
        The returned object can be itself called like a function.  Given a lead
        number, it returns a 2d NumPy array that contains the wave function
        within the scattering region due to each incoming mode of the given
        lead.  Index 0 is the mode number, index 1 is the orbital number.

        The modes appear in the same order as the negative velocity modes in
        `kwant.physics.PropagatingModes`. In Kwant's convention leads are attached
        so that their translational symmetry points *away* from the scattering
        region::

             left lead    SR   right lead
             /---------\ /---\ /---------\
             ...-3-2-1-0-X-X-X-0-1-2-3-...

        This means that incoming modes (coming from infinity towards the
        scattering region) have *negative* velocity with respect to the
        lead's symmetry direction.

        Examples
        --------
        >>> wf = kwant.solvers.default.wave_function(some_syst, some_energy)
        >>> wfs_of_lead_2 = wf(2)

        """
        return WaveFunction(self, sys, energy, args, check_hermiticity, params)


class WaveFunction:
    def __init__(self, solver, sys, energy, args, check_hermiticity, params):
        syst = sys  # ensure consistent naming across function bodies
        ensure_isinstance(syst, system.System)
        for lead in syst.leads:
            if not hasattr(lead, 'modes'):
                # TODO: figure out what to do with self-energies.
                msg = ('Wave functions for leads with only self-energy'
                       ' are not available yet.')
                raise NotImplementedError(msg)
        linsys = solver._make_linear_sys(syst, range(len(syst.leads)), energy,
                                         args, check_hermiticity,
                                         params=params)[0]
        self.solve = solver._solve_linear_sys
        self.rhs = linsys.rhs
        self.factorized_h = solver._factorized(linsys.lhs)
        self.num_orb = linsys.num_orb

    def __call__(self, lead):
        result = self.solve(self.factorized_h, self.rhs[lead],
                            slice(self.num_orb))
        return result.transpose()


class BlockResult(metaclass=abc.ABCMeta):
    """
    ABC for a linear system solution with variable grouping.
    """

    def __init__(self, data, lead_info, out_leads, in_leads, sizes,
                 current_conserving=False):
        self.data = data
        self.lead_info = lead_info
        self.out_leads = out_leads = list(out_leads)
        self.in_leads = in_leads = list(in_leads)
        self.sizes = sizes = np.array(sizes)
        self.current_conserving = current_conserving
        self.in_offsets = np.zeros(len(in_leads) + 1, int)
        self.in_offsets[1 :] = np.cumsum(self.sizes[in_leads])
        self.out_offsets = np.zeros(len(self.out_leads) + 1, int)
        self.out_offsets[1 :] = np.cumsum(self.sizes[out_leads])

    def block_coords(self, lead_out, lead_in):
        """
        Return slices corresponding to the block from lead_in to lead_out.
        """
        return self.out_block_coords(lead_out), self.in_block_coords(lead_in)

    def out_block_coords(self, lead_out):
        """Return a slice with the rows in the block corresponding to lead_out.
        """
        lead_out = self.out_leads.index(lead_out)
        return slice(self.out_offsets[lead_out],
                     self.out_offsets[lead_out + 1])

    def in_block_coords(self, lead_in):
        """
        Return a slice with the columns in the block corresponding to lead_in.
        """
        lead_in = self.in_leads.index(lead_in)
        return slice(self.in_offsets[lead_in],
                     self.in_offsets[lead_in + 1])

    def submatrix(self, lead_out, lead_in):
        """Return the matrix elements from lead_in to lead_out."""
        return self.data[self.block_coords(lead_out, lead_in)]

    @abc.abstractmethod
    def num_propagating(self, lead):
        """Return the number of propagating modes in the lead."""
        pass

    @abc.abstractmethod
    def _transmission(self, lead_out, lead_in):
        """Return transmission from lead_in to lead_out.

        The calculation is expected to be done directly from the corresponding
        matrix elements.
        """
        pass

    def transmission(self, lead_out, lead_in):
        """Return transmission from lead_in to lead_out.

        If the option ``current_conserving`` has been enabled for this object,
        this method will deduce missing transmission values whenever possible.

        Current conservation is enabled by default for objects returned by
        `~kwant.solvers.default.smatrix` and
        `~kwant.solvers.default.greens_function` whenever the Hamiltonian has
        been verified to be Hermitian (option ``check_hermiticity``, enabled by
        default).

        """
        chosen = [lead_out, lead_in]
        present = self.out_leads, self.in_leads
        # OK means: chosen (outgoing, incoming) lead is present.
        ok = [int(c in p) for c, p in zip(chosen, present)]
        if all(ok):
            return self._transmission(lead_out, lead_in)

        if self.current_conserving:
            all_but_one = len(self.lead_info) - 1
            s_t = self._transmission
            if sum(ok) == 1:
                # Calculate the transmission element by summing over a single
                # row or column.
                sum_dim, ok_dim = ok
                if len(present[sum_dim]) == all_but_one:
                    return (self.num_propagating(chosen[ok_dim]) - sum(
                        s_t(*chosen) for chosen[sum_dim] in present[sum_dim]))
            elif all(len(p) == all_but_one for p in present):
                # Calculate the transmission element at the center of a single
                # missing row and a single missing column.
                return (sum(s_t(o, i) - self.num_propagating(o) if o == i else 0
                            for o, i in product(*present)))

        raise ValueError("Insufficient matrix elements to compute "
                         "transmission({0}, {1}).".format(*chosen))

    def conductance_matrix(self):
        """Return the conductance matrix.

        This is the matrix :math:`C` such that :math:`I = CV` where :math:`I`
        and :math:`V` are, respectively, the vectors of currents and voltages
        for each lead.

        This matrix is useful for calculating non-local resistances.  See
        Section 2.4 of the book by S. Datta.
        """
        n = len(self.lead_info)
        rn = range(n)
        result = np.array([[-self.transmission(i, j) if i != j else 0
                            for j in rn] for i in rn])
        # Set the diagonal elements such that the sums of columns are zero.
        result.flat[::n+1] = -result.sum(axis=0)
        return result


class SMatrix(BlockResult):
    """A scattering matrix.

    Transport properties can be easily accessed using the
    `~SMatrix.transmission` method (don't be fooled by the name,
    it can also compute reflection, which is just transmission from one
    lead back into the same lead.)

    `SMatrix` however also allows for a more direct access to the result: The
    data stored in `SMatrix` is a scattering matrix with respect to lead modes
    and these modes themselves. The details of this data can be directly
    accessed through the instance variables `data` and `lead_info`. Subblocks
    of data corresponding to particular leads are conveniently obtained by
    `~SMatrix.submatrix`.

    `SMatrix` also respects the conservation laws present in the lead, such as
    spin conservation, if they are declared during system construction. If
    queried with length-2 sequence the first number is the number of the lead,
    and the second number is the index of the corresponding conserved
    quantity. For example ``smatrix.transmission((1, 3), (0, 2))`` is
    transmission from block 2 of the conserved quantity in lead 0 to the block
    3 of the conserved quantity in lead 1.

    Attributes
    ----------
    data : NumPy array
        a matrix containing all the requested matrix elements of the scattering
        matrix.
    lead_info : list of data
        a list containing `kwant.physics.PropagatingModes` for each lead.
    out_leads, in_leads : sequence of integers
        indices of the leads where current is extracted (out) or injected
        (in). Only those are listed for which SMatrix contains the
        calculated result.
    """

    def __init__(self, data, lead_info, out_leads, in_leads,
                 current_conserving=False):
        sizes = [len(i.momenta) // 2 for i in lead_info]
        super().__init__(data, lead_info, out_leads, in_leads,
                                      sizes, current_conserving)
        # in_offsets marks beginnings and ends of blocks in the scattering
        # matrix corresponding to the in leads. The end of the block
        # corresponding to lead i is taken as the beginning of the block of
        # i+1.  Same for out leads. For each lead block of the scattering
        # matrix, we want to pick out the computed blocks of the conservation
        # law.  The offsets of these symmetry blocks are stored in
        # block_offsets, for all leads.  List of lists containing the sizes of
        # symmetry blocks in all leads. leads_block_sizes[i][j] is the number
        # of left or right moving modes in symmetry block j of lead
        # i. len(leads_block_sizes[i]) is the number of blocks in lead i.
        leads_block_sizes = []
        for info in self.lead_info:
            # If a lead does not have block structure, append None.
            leads_block_sizes.append(getattr(info, 'block_nmodes', None))
        self.leads_block_sizes = leads_block_sizes
        block_offsets = []
        for lead_block_sizes in self.leads_block_sizes: # Cover all leads
            if lead_block_sizes is None:
                block_offsets.append(lead_block_sizes)
            else:
                block_offset = np.zeros(len(lead_block_sizes) + 1, int)
                block_offset[1:] = np.cumsum(lead_block_sizes)
                block_offsets.append(block_offset)
        # Symmetry block offsets for all leads - or None if lead does not have
        # blocks.
        self.block_offsets = block_offsets
        # Pick out symmetry block offsets for in and out leads
        self.in_block_offsets = \
                np.array(self.block_offsets)[list(self.in_leads)]
        self.out_block_offsets = \
                np.array(self.block_offsets)[list(self.out_leads)]
        # Block j of in lead i starts at in_block_offsets[i][j]

    def out_block_coords(self, lead_out):
        """Return a slice with the rows in the block corresponding to lead_out.
        """
        if isinstance(lead_out, Integral):
            lead_out = self.out_leads.index(lead_out)
            return slice(self.out_offsets[lead_out],
                         self.out_offsets[lead_out + 1])
        else:
            lead_ind, block_ind = lead_out
            lead_ind = self.out_leads.index(lead_ind)
            return slice(self.out_offsets[lead_ind] +
                         self.out_block_offsets[lead_ind][block_ind],
                         self.out_offsets[lead_ind] +
                         self.out_block_offsets[lead_ind][block_ind + 1])

    def in_block_coords(self, lead_in):
        """
        Return a slice with the columns in the block corresponding to lead_in.
        """
        if isinstance(lead_in, Integral):
            lead_in = self.in_leads.index(lead_in)
            return slice(self.in_offsets[lead_in],
                     self.in_offsets[lead_in + 1])
        else:
            lead_ind, block_ind = lead_in
            lead_ind = self.in_leads.index(lead_ind)
            return slice(self.in_offsets[lead_ind] +
                         self.in_block_offsets[lead_ind][block_ind],
                         self.in_offsets[lead_ind] +
                         self.in_block_offsets[lead_ind][block_ind + 1])

    def _transmission(self, lead_out, lead_in):
        return np.linalg.norm(self.submatrix(lead_out, lead_in)) ** 2

    def transmission(self, lead_out, lead_in):
        """Return transmission from lead_in to lead_out.

        If `lead_in` or `lead_out` are a length-2 sequence, the first number is
        the number of the lead, and the second number indexes the eigenvalue of
        the conserved quantity in that lead (e.g. spin) if it was specified.
        """
        # If transmission is between leads and not subblocks, we can use
        # unitarity (like BlockResult does).
        if isinstance(lead_in, Integral) and isinstance(lead_out, Integral):
            return super().transmission(lead_out, lead_in)
        else:
            return self._transmission(lead_out, lead_in)

    def num_propagating(self, lead):
        """Return the number of propagating modes in the lead."""
        return self.sizes[lead]

    def __repr__(self):
        return ("SMatrix(data=%r, lead_info=%r, out_leads=%r, in_leads=%r)" %
                (self.data, self.lead_info, self.out_leads, self.in_leads))


class GreensFunction(BlockResult):
    """
    Retarded Green's function.

    Transport properties can be easily accessed using the
    `~GreensFunction.transmission` method (don't be fooled by the name, it can
    also compute reflection, which is just transmission from one lead back into
    the same lead).

    `GreensFunction` however also allows for a more direct access to the
    result: The data stored in `GreensFunction` is the real-space Green's
    function. The details of this data can be directly accessed through the
    instance variables `data` and `lead_info`. Subblocks of data corresponding
    to particular leads are conveniently obtained by
    `~GreensFunction.submatrix`.

    Attributes
    ----------
    data : NumPy array
        a matrix containing all the requested matrix elements of Green's
        function.
    lead_info : list of matrices
        a list with self-energies of each lead.
    out_leads, in_leads : sequence of integers
        indices of the leads where current is extracted (out) or injected
        (in). Only those are listed for which SMatrix contains the
        calculated result.
    """

    def __init__(self, data, lead_info, out_leads, in_leads,
                 current_conserving=False):
        sizes = [i.shape[0] for i in lead_info]
        super().__init__(
            data, lead_info, out_leads, in_leads, sizes, current_conserving)

    def num_propagating(self, lead):
        """Return the number of propagating modes in the lead."""
        gamma = 1j * (self.lead_info[lead] -
                      self.lead_info[lead].conj().T)

        # The number of channels is given by the number of
        # nonzero eigenvalues of Gamma
        # rationale behind the threshold from
        # Golub; van Loan, chapter 5.5.8
        eps = np.finfo(gamma.dtype).eps * 1000
        return np.sum(np.linalg.eigvalsh(gamma) >
                      eps * np.linalg.norm(gamma, np.inf))

    def _a_ttdagger_a_inv(self, lead_out, lead_in):
        """Return t * t^dagger in a certain basis."""
        gf = self.submatrix(lead_out, lead_in)
        factors = []
        for lead, gf2 in ((lead_out, gf), (lead_in, gf.conj().T)):
            possible_se = self.lead_info[lead]
            factors.append(1j * (possible_se - possible_se.conj().T))
            factors.append(gf2)
        return reduce(np.dot, factors)

    def _transmission(self, lead_out, lead_in):
        attdagainv = self._a_ttdagger_a_inv(lead_out, lead_in)
        result = np.trace(attdagainv).real
        if lead_out == lead_in:
            # For reflection we have to be more careful
            gamma = 1j * (self.lead_info[lead_in] -
                          self.lead_info[lead_in].conj().T)
            gf = self.submatrix(lead_out, lead_in)

            # The number of channels is given by the number of
            # nonzero eigenvalues of Gamma
            # rationale behind the threshold from
            # Golub; van Loan, chapter 5.5.8
            eps = np.finfo(gamma.dtype).eps * 1000
            N = np.sum(np.linalg.eigvalsh(gamma) >
                       eps * np.linalg.norm(gamma, np.inf))

            result += 2 * np.trace(np.dot(gamma, gf)).imag + N

        return result

    def __repr__(self):
        return ("GreensFunction(data=%r, lead_info=%r, "
                "out_leads=%r, in_leads=%r)" %
                (self.data, self.lead_info, self.out_leads, self.in_leads))



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['gen_eig']

from . import lapack


def gen_eig(a, b, left=False, right=True, overwrite_ab=False):
    """Compute the eigenvalues and -vectors of the matrix pencil (a,b), i.e. of
    the generalized (unsymmetric) eigenproblem a v = lambda b v where a and b
    are square (unsymmetric) matrices, v the eigenvector and lambda the
    eigenvalues.

    The eigenvalues are returned as numerator alpha and denominator beta,
    i.e. lambda = alpha/beta. This is advantageous, as lambda can be infinity
    which is well-defined in this case as beta = 0.

    Parameters
    ----------
    a : array, shape (M, M)
    b : array, shape (M, M)
        `a` and `b` are the two matrices defining the generalized eigenproblem
    left : boolean
        Whether to calculate and return left eigenvectors
    right : boolean
        Whether to calculate and return right eigenvectors

    overwrite_ab : boolean
        Whether to overwrite data in `a` and `b` (may improve performance)

    Returns
    -------
    alpha : complex array, shape (M,)
    beta : real or complex array, shape (M,)
        The eigenvalues in the form ``alpha/beta``

    (if left == True)
    vl : double or complex array, shape (M, M)
        The left eigenvector corresponding to the eigenvalue
        ``alpha[i]/beta[i]`` is the column ``vl[:,i]``.

    (if right == True)
    vr : double or complex array, shape (M, M)
        The right eigenvector corresponding to the eigenvalue
        ``alpha[i]/beta[i]`` is the column ``vr[:,i]``.
    """
    a, b = lapack.prepare_for_lapack(overwrite_ab, a, b)
    return lapack.ggev(a, b, left, right)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['smatrix', 'ldos', 'wave_function', 'greens_function', 'options',
           'Solver']

import numpy as np
from . import common
from ..linalg import mumps


class Solver(common.SparseSolver):
    """Sparse Solver class based on the sparse direct solver MUMPS."""

    lhsformat = 'coo'
    rhsformat = 'csc'

    def __init__(self):
        self.nrhs = self.ordering = self.sparse_rhs = None
        self.reset_options()

    def reset_options(self):
        """Set the options to default values.  Return the old options."""
        return self.options(nrhs=6, ordering='kwant_decides', sparse_rhs=False)

    def options(self, nrhs=None, ordering=None, sparse_rhs=None):
        """
        Modify some options.  Return the old options.

        Parameters
        ----------
        nrhs : number
            number of right hand sides that should be solved simultaneously. A
            value around 5-10 gives optimal performance on many machines.  If
            memory is an issue, it can be set to 1, to minimize memory usage
            (at the cost of slower performance). Default value is 6.
        ordering : string
            one of the ordering methods supported by the MUMPS solver (see
            ``kwant.linalg.mumps``. The availability of certain orderings
            depends on the MUMPS installation.), or 'kwant_decides'.  If
            ``ordering=='kwant_decides'``, the ordering that typically gives
            the best performance is chosen from the available ones.  One can
            also defer the choice of ordering to MUMPS by specifying 'auto', in
            some cases MUMPS however chooses poorly.

            The choice of ordering can significantly influence the performance
            and memory impact of the solve phase. Typically the nested
            dissection orderings 'metis' and 'scotch' are most suited for
            physical systems. Default is 'kwant_decides'
        sparse_rhs : True or False
            whether to use a sparse right hand side in the solve phase of
            MUMPS. Preliminary tests have not shown a significant performance
            increase when this feature is used, but this needs more looking
            into. Default value is False.

        Returns
        -------
        old_options: dict
            dictionary containing the previous options.

        Notes
        -----
        Thanks to this method returning the old options as a dictionary it is
        easy to change some options temporarily:

        >>> saved_options = kwant.solvers.mumps.options(nrhs=12)
        >>> some_code()
        >>> kwant.solvers.mumps.options(**saved_options)
        """

        old_opts = {'nrhs': self.nrhs,
                    'ordering': self.ordering,
                    'sparse_rhs': self.sparse_rhs}

        if nrhs is not None:
            if nrhs < 1 and int(nrhs) != nrhs:
                raise ValueError("nrhs must be an integer bigger than zero")
            nrhs = int(nrhs)
            self.nrhs = nrhs

        if ordering is not None:
            if ordering == 'kwant_decides':
                # Choose what is considered to be the best ordering.
                sorted_orderings = [order
                                    for order in ['metis', 'scotch', 'auto']
                                    if order in mumps.possible_orderings()]
                ordering = sorted_orderings[0]
            elif ordering not in mumps.orderings:
                raise ValueError("Invalid ordering: " + ordering)
            self.ordering = ordering

        if sparse_rhs is not None:
            self.sparse_rhs = bool(sparse_rhs)

        return old_opts

    def _factorized(self, a):
        inst = mumps.MUMPSContext()
        inst.factor(a, ordering=self.ordering)
        return inst

    def _solve_linear_sys(self, factorized_a, b, kept_vars):
        if b.shape[1] == 0:
            return b[kept_vars]

        solve = factorized_a.solve
        sols = []

        for j in range(0, b.shape[1], self.nrhs):
            tmprhs = b[:, j:min(j + self.nrhs, b.shape[1])]

            if not self.sparse_rhs:
                tmprhs = tmprhs.toarray()
            sols.append(solve(tmprhs)[kept_vars, :])

        return np.concatenate(sols, axis=1)


default_solver = Solver()

smatrix = default_solver.smatrix
greens_function = default_solver.greens_function
ldos = default_solver.ldos
wave_function = default_solver.wave_function
options = default_solver.options
reset_options = default_solver.reset_options



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['smatrix', 'greens_function', 'ldos', 'wave_function', 'Solver']

import numpy as np
import scipy.sparse as sp
from . import common

# Note: previous code would have failed if UMFPACK was provided by scikit
import scipy.sparse.linalg.dsolve.linsolve as linsolve

# check if we are actually using UMFPACK or rather SuperLU
# TODO: remove the try (only using the except clause) once we depend on
# scipy >= 0.14.0.
try:
    uses_umfpack = linsolve.isUmfpack
except AttributeError:
    uses_umfpack = linsolve.useUmfpack

if uses_umfpack:
    umfpack = linsolve.umfpack

if uses_umfpack:
    def factorized(A, piv_tol=1.0, sym_piv_tol=1.0):
        """
        Return a fuction for solving a sparse linear system, with A
        pre-factorized.

        Parameters
        ----------
        A : csc_matrix
            matrix to be factorized
        piv_tol : float, 0 <= piv_tol <= 1.0
        sym_piv_tol : float, 0 <= piv_tol <= 1.0
            thresholds used by UMFPACK for pivoting. 0 means no pivoting, 1.0
            means full pivoting as in dense matrices (guaranteeing stability,
            but reducing possibly sparsity). Defaults of UMFPACK are 0.1 and
            0.001 respectively. Whether piv_tol or sym_piv_tol are used is
            decided internally by UMFPACK, depending on whether the matrix is
            "symmetric" enough.

        Examples
        --------
        solve = factorized(A) # Makes LU decomposition.
        x1 = solve(rhs1) # Uses the LU factors.
        x2 = solve(rhs2) # Uses again the LU factors.
        """

        if not sp.isspmatrix_csc(A):
            A = sp.csc_matrix(A)

        A.sort_indices()
        A = A.asfptype()  # upcast to a floating point format

        if A.dtype.char not in 'dD':
            raise ValueError("convert matrix data to double, please, using"
                             " .astype()")

        family = {'d': 'di', 'D': 'zi'}
        umf = umfpack.UmfpackContext(family[A.dtype.char])

        # adjust pivot thresholds
        umf.control[umfpack.UMFPACK_PIVOT_TOLERANCE] = piv_tol
        umf.control[umfpack.UMFPACK_SYM_PIVOT_TOLERANCE] = sym_piv_tol

        # Make LU decomposition.
        umf.numeric(A)

        def solve(b):
            return umf.solve(umfpack.UMFPACK_A, A, b, autoTranspose=True)

        return solve
else:
    # no UMFPACK found. SuperLU is being used, but usually abysmally slow
    # (SuperLu is not bad per se, somehow the SciPy version isn't good).
    # Since scipy doesn't include UMFPACK anymore due to software rot,
    # there is no warning here.
    factorized = linsolve.factorized


class Solver(common.SparseSolver):
    "Sparse Solver class based on the sparse direct solvers provided by SciPy."
    lhsformat = 'csc'
    rhsformat = 'csc'
    nrhs = 1

    def _factorized(self, a):
        a = sp.csc_matrix(a)
        return factorized(a)

    def _solve_linear_sys(self, factorized_a, b, kept_vars):
        if b.shape[1] == 0:
            return b[kept_vars]

        sols = []
        vec = np.empty(b.shape[0], complex)
        for j in range(b.shape[1]):
            vec[:] = b[:, j].toarray().flatten()
            sols.append(factorized_a(vec)[kept_vars])

        return np.asarray(sols).transpose()


default_solver = Solver()

smatrix = default_solver.smatrix
greens_function = default_solver.greens_function
ldos = default_solver.ldos
wave_function = default_solver.wave_function



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import math
import numpy as np
from .. import system
from .._common import ensure_isinstance

__all__ = ['Bands']


class Bands:
    """
    Class of callable objects for the computation of energy bands.

    Parameters
    ----------
    sys : `kwant.system.InfiniteSystem`
        The low level infinite system for which the energies are to be
        calculated.
    args : tuple, defaults to empty
        Positional arguments to pass to the ``hamiltonian`` method.
        Mutually exclusive with 'params'.
    params : dict, optional
        Dictionary of parameter names and their values. Mutually exclusive
        with 'args'.

    Notes
    -----
    An instance of this class can be called like a function.  Given a momentum
    (currently this must be a scalar as all infinite systems are quasi-1-d), it
    returns a NumPy array containing the eigenenergies of all modes at this
    momentum

    Examples
    --------
    >>> bands = kwant.physics.Bands(some_syst)
    >>> momenta = numpy.linspace(-numpy.pi, numpy.pi, 101)
    >>> energies = [bands(k) for k in momenta]
    >>> pyplot.plot(momenta, energies)
    >>> pyplot.show()
    """

    def __init__(self, sys, args=(), *, params=None):
        syst = sys
        ensure_isinstance(syst, system.InfiniteSystem)
        self.ham = syst.cell_hamiltonian(args, params=params)
        if not np.allclose(self.ham, self.ham.T.conj()):
            raise ValueError('The cell Hamiltonian is not Hermitian.')
        hop = syst.inter_cell_hopping(args, params=params)
        self.hop = np.empty(self.ham.shape, dtype=complex)
        self.hop[:, : hop.shape[1]] = hop
        self.hop[:, hop.shape[1]:] = 0

    def __call__(self, k):
        # Note: Equation to solve is
        #       (V^\dagger e^{ik} + H + V e^{-ik}) \psi = E \psi
        mat = self.hop * complex(math.cos(k), -math.sin(k))
        mat += mat.conjugate().transpose() + self.ham
        return np.sort(np.linalg.eigvalsh(mat).real)



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.


import numpy as np
from kwant.linalg import lll
from kwant._common import ensure_rng

def test_lll():
    rng = ensure_rng(1)
    for i in range(50):
        x = rng.randint(4) + 1
        mat = rng.randn(x, x + rng.randint(2))
        c = 1.34 + .5 * rng.random_sample()
        reduced_mat, coefs = lll.lll(mat)
        assert lll.is_c_reduced(reduced_mat, c)
        assert np.allclose(np.dot(mat.T, coefs), reduced_mat.T)


def test_cvp():
    rng = ensure_rng(0)
    for i in range(1, 5):
        for j in range(i, 5):
            mat = rng.randn(i, j)
            mat = lll.lll(mat)[0]
            for k in range(4):
                point = 50 * rng.randn(j)
                assert np.array_equal(lll.cvp(point, mat, 10)[:3],
                                      lll.cvp(point, mat, 3))



# -*- coding: utf-8 -*-
# Copyright 2011-2016 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

import math
import numpy as np
import scipy
import scipy.sparse.linalg as sla
import scipy.fftpack as fft

from . import system
from ._common import ensure_rng
from .operator import _LocalOperator

__all__ = ['SpectralDensity']


class SpectralDensity:
    """Calculate the spectral density of an operator.

    This class makes use of the kernel polynomial
    method (KPM), presented in [1]_, to obtain the spectral density
    :math:`ρ_A(e)`, as a function of the energy :math:`e`, of some
    operator :math:`A` that acts on a kwant system or a Hamiltonian.
    In general

    .. math::
       ρ_A(e) = ρ(e) A(e),

    where :math:`ρ(e) = \sum_{k=0}^{D-1} δ(E-E_k)` is the density of
    states, and :math:`A(e)` is the expectation value of :math:`A` for
    all the eigenstates with energy :math:`e`.

    Parameters
    ----------
    hamiltonian : `~kwant.system.FiniteSystem` or matrix Hamiltonian
        If a system is passed, it should contain no leads.
    params : dict, optional
        Additional parameters to pass to the Hamiltonian and operator.
    operator : operator, dense matrix, or sparse matrix, optional
        Operator for which the spectral density will be evaluated. If
        it is callable, the ``densities`` at each energy will have the
        dimension of the result of `operator(bra, ket)`. If it has a
        ``dot`` method, such as ``numpy.ndarray`` and
        ``scipy.sparse.matrices``, the densities will be scalars.
        If no operator is provided, the density of states is calculated
        with a faster algorithm.
    num_vectors : positive int, default: 10
        Number of random vectors for the KPM method.
    num_moments : positive int, default: 100
        Number of moments, order of the KPM expansion. Mutually exclusive
        with 'energy_resolution'.
    energy_resolution : positive float, optional
        The resolution in energy of the KPM approximation to the spectral
        density. Mutually exclusive with 'num_moments'.
    vector_factory : function, optional
        The user defined function ``f(n)`` generates random vectors of
        length ``n`` that will be used in the algorithm.
        If not provided, random phase vectors are used.
        The default random vectors are optimal for most cases, see the
        discussions in [1]_ and [2]_.
    bounds : pair of floats, optional
        Lower and upper bounds for the eigenvalue spectrum of the system.
        If not provided, they are computed.
    eps : positive float, default: 0.05
        Parameter to ensure that the rescaled spectrum lies in the
        interval ``(-1, 1)``; required for stability.
    rng : seed, or random number generator, optional
        Random number generator used for the calculation of the spectral
        bounds, and to generate random vectors (if ``vector_factory`` is
        not provided). If not provided, numpy's rng will be used; if it
        is an Integer, it will be used to seed numpy's rng, and if it is
        a random number generator, this is the one used.

    Notes
    -----
    When passing an ``operator`` defined in `~kwant.operator`, the
    result of ``operator(bra, ket)`` depends on the attribute ``sum``
    of such operator. If ``sum=True``, densities will be scalars, that
    is, total densities of the system. If ``sum=False`` the densities
    will be arrays of the length of the system, that is, local
    densities.

    .. [1] `Rev. Mod. Phys., Vol. 78, No. 1 (2006)
       <https://arxiv.org/abs/cond-mat/0504627>`_.
    .. [2] `Phys. Rev. E 69, 057701 (2004)
       <https://arxiv.org/abs/cond-mat/0401202>`_

    Examples
    --------
    In the following example, we will obtain the density of states of a
    graphene sheet, defined as a honeycomb lattice with first nearest
    neighbors coupling.

    We start by importing kwant and defining a
    `~kwant.system.FiniteSystem`,

    >>> import kwant
    ...
    >>> def circle(pos):
    ...     x, y = pos
    ...     return x**2 + y**2 < 100
    ...
    >>> lat = kwant.lattice.honeycomb()
    >>> syst = kwant.Builder()
    >>> syst[lat.shape(circle, (0, 0))] = 0
    >>> syst[lat.neighbors()] = 1

    and after finalizing the system, create an instance of
    `~kwant.kpm.SpectralDensity`

    >>> fsyst = syst.finalized()
    >>> rho = kwant.kpm.SpectralDensity(fsyst)

    The ``energies`` and ``densities`` can be accessed with

    >>> energies, densities = rho()

    or

    >>> energies, densities = rho.energies, rho.densities

    Attributes
    ----------
    energies : array of floats
        Array of sampling points with length ``2 * num_moments`` in
        the range of the spectrum.
    densities : array of floats
        Spectral density of the ``operator`` evaluated at the energies.
    """

    def __init__(self, hamiltonian, params=None, operator=None,
                 num_vectors=10, num_moments=None, energy_resolution=None,
                 vector_factory=None, bounds=None, eps=0.05, rng=None):

        if num_moments and energy_resolution:
            raise TypeError("either 'num_moments' or 'energy_resolution' "
                            "must be provided.")

        rng = ensure_rng(rng)
        # self.eps ensures that the rescaled Hamiltonian has a
        # spectrum strictly in the interval (-1,1).
        self.eps = eps

        # Normalize the format of 'ham'
        if isinstance(hamiltonian, system.System):
            hamiltonian = hamiltonian.hamiltonian_submatrix(params=params,
                                                            sparse=True)
        try:
            hamiltonian = scipy.sparse.csr_matrix(hamiltonian)
        except Exception:
            raise ValueError("'hamiltonian' is neither a matrix "
                             "nor a Kwant system.")

        # Normalize 'operator' to a common format.
        if operator is None:
            self.operator = None
        elif isinstance(operator, _LocalOperator):
            self.operator = operator.bind(params=params)
        elif callable(operator):
            self.operator = operator
        elif hasattr(operator, 'dot'):
            operator = scipy.sparse.csr_matrix(operator)
            self.operator = lambda bra, ket: np.vdot(bra, operator.dot(ket))
        else:
            raise ValueError('Parameter `operator` has no `.dot` '
                             'attribute and is not callable.')

        self._vector_factory = vector_factory or \
            (lambda n: np.exp(2j * np.pi * rng.random_sample(n)))
        # store this vector for reproducibility
        self._v0 = np.exp(2j * np.pi * rng.random_sample(hamiltonian.shape[0]))
        self._rand_vect_list = []
        # Hamiltonian rescaled as in Eq. (24)
        self.hamiltonian, (self._a, self._b) = _rescale(hamiltonian,
                                                        eps=self.eps,
                                                        v0=self._v0,
                                                        bounds=bounds)
        self.bounds = (self._b - self._a, self._b + self._a)

        if energy_resolution:
            num_moments = math.ceil((1.6 * self._a) / energy_resolution)
        elif num_moments is None:
            num_moments = 100

        must_be_positive_int = ['num_vectors', 'num_moments']
        for var in must_be_positive_int:
            val = locals()[var]
            if val <= 0 or val != int(val):
                raise ValueError('{} must be a positive integer'.format(var))
        if eps <= 0:
            raise ValueError('eps must be positive')

        for r in range(num_vectors):
            self._rand_vect_list.append(
                self._vector_factory(self.hamiltonian.shape[0]))
        self._last_two_alphas = [0.] * num_vectors
        self._moments_list = [0.] * num_vectors

        self.num_moments = num_moments
        self.num_vectors = 0  # new random vectors will be used
        self._update_moments_list(self.num_moments, num_vectors)
        self.num_vectors = num_vectors

        moments = self._moments()
        xk_rescaled, rho, self._gammas = _calc_fft_moments(
            moments, 2 * self.num_moments)
        self.energies = xk_rescaled * self._a + self._b
        self.densities = rho

    def __call__(self, energy=None):
        """Return the spectral density evaluated at ``energy``.

        Parameters
        ----------
        energy : float or sequence of float, optional

        Returns
        -------
        float, if ``energy`` is float, array of float if ``energy``
        is a sequence, a tuple (energies, densities) if
        ``energy`` was not provided.

        Notes
        -----
        If ``energy`` is not provided, then the densities are obtained
        by Fast Fourier Transform of the Chebyshev moments.
        """
        if energy is None:
            return self.energies, self.densities
        else:
            energy = np.asarray(energy, dtype=complex)
            rescaled_energy = (energy - self._b) / self._a
            g_e = (np.pi * np.sqrt(1 - rescaled_energy)
                   * np.sqrt(1 + rescaled_energy))

            moments = self._moments()

            m = np.arange(self.num_moments)

            kernel = ((self.num_moments - m + 1) *
                      np.cos(np.pi * m/(self.num_moments + 1)) +
                      np.sin(np.pi * m/(self.num_moments + 1)) /
                      np.tan(np.pi/(self.num_moments + 1)))
            kernel = kernel / (self.num_moments + 1)

            # transposes handle the case where operators have vector outputs
            coef_cheb = np.transpose(moments.transpose() * kernel)
            coef_cheb[1:] = 2 * coef_cheb[1:]

            return np.transpose(np.polynomial.chebyshev.chebval(
                    rescaled_energy, coef_cheb) / g_e).real

    def integrate(self, distribution_function=None):
        """Returns the total spectral density.

        Returns the integral over the whole spectrum with an optional
        distribution function. ``distribution_function`` should be able
        to take arrays as input. Defined using Gauss-Chebyshev
        integration.
        """
        # This factor divides the sum to normalize the Gauss integral
        # and rescales the integral back with ``self._a`` to normal
        # scale.
        factor = self._a / (2 * self.num_moments)
        if distribution_function is None:
            rho = self._gammas
        else:
            # The evaluation of the distribution function should be at
            # the energies without rescaling.
            distribution_array = distribution_function(self.energies)
            rho = np.transpose(self._gammas.transpose() * distribution_array)
        return factor * np.sum(rho, axis=0)

    def add_moments(self, num_moments=None, *, energy_resolution=None):
        """Increase the number of Chebyshev moments.

        Parameters
        ----------
        num_moments: positive int
            The number of Chebyshev moments to add. Mutually
            exclusive with 'energy_resolution'.
        energy_resolution: positive float, optional
            Features wider than this resolution are visible
            in the spectral density. Mutually exclusive with
            'num_moments'.
        """
        if not ((num_moments is None) ^ (energy_resolution is None)):
            raise TypeError("either 'num_moments' or 'energy_resolution' "
                            "must be provided.")

        if energy_resolution:
            if energy_resolution <= 0:
                raise ValueError("'energy_resolution' must be positive"
                                 .format(energy_resolution))
            # factor of 1.6 comes from the fact that we use the
            # Jackson kernel when calculating the FFT, which has
            # maximal slope π/2. Rounding to 1.6 ensures that the
            # energy resolution is sufficient.
            present_resolution = self._a * 1.6 / self.num_moments
            if present_resolution < energy_resolution:
                raise ValueError('Energy resolution is already smaller '
                                 'than the requested resolution')
            num_moments = math.ceil((1.6 * self._a) / energy_resolution)

        if (num_moments is None or num_moments <= 0
            or num_moments != int(num_moments)):
            raise ValueError("'num_moments' must be a positive integer")

        self._update_moments_list(self.num_moments + num_moments,
                                  self.num_vectors)
        self.num_moments += num_moments

        # recalculate quantities derived from the moments
        moments = self._moments()
        xk_rescaled, rho, self._gammas = _calc_fft_moments(
            moments, 2 * self.num_moments)
        self.energies = xk_rescaled * self._a + self._b
        self.densities = rho

    def add_vectors(self, num_vectors):
        """Increase the number of random vectors.

        Parameters
        ----------
        num_vectors: positive int
            The number of random vectors to add.
        """
        if num_vectors <= 0 or num_vectors != int(num_vectors):
            raise ValueError("'num_vectors' must be a positive integer")
        for r in range(num_vectors):
            self._rand_vect_list.append(
                self._vector_factory(self.hamiltonian.shape[0]))
        self._moments_list.extend([0.] * num_vectors)
        self._last_two_alphas.extend([0.] * num_vectors)
        self._update_moments_list(self.num_moments,
                                  self.num_vectors + num_vectors)
        self.num_vectors += num_vectors

        # recalculate quantities derived from the moments
        moments = self._moments()
        xk_rescaled, rho, self._gammas = _calc_fft_moments(
            moments, 2 * self.num_moments)
        self.energies = xk_rescaled * self._a + self._b
        self.densities = rho

    def _moments(self):
        # sum moments of all random vectors
        moments = np.sum(np.asarray(self._moments_list).real, axis=0)
        # divide by the number of random vectors
        moments /= self.num_vectors
        # divide by scale factor to reflect the integral rescaling
        return moments / self._a

    def _update_moments_list(self, n_moments, n_rand):
        """Calculate the Chebyshev moments of an operator's spectral
        density.

        The algorithm is based on the KPM method as depicted in `Rev.
        Mod. Phys., Vol. 78, No. 1 (2006)
        <https://arxiv.org/abs/cond-mat/0504627>`_.

        Parameters
        ----------
        n_moments : integer
            Number of Chebyshev moments.
        n_rand : integer
            Number of random vectors used for sampling.
        """

        if self.num_vectors == n_rand:
            r_start = 0
            new_rand_vect = 0
        elif self.num_vectors < n_rand:
            r_start = self.num_vectors
            new_rand_vect = n_rand - self.num_vectors
        else:
            raise ValueError('Cannot decrease number of random vectors')

        if n_moments == self.num_moments:
            m_start = 2
            new_moments = 0
            if new_rand_vect == 0:
                # nothing new to calculate
                return
        else:
            new_moments = n_moments - self.num_moments
            m_start = self.num_moments
            if new_moments < 0:
                raise ValueError('Cannot decrease number of moments')

            if new_rand_vect != 0:
                raise ValueError("Only 'num_moments' *or* 'num_vectors' "
                                 "may be updated at a time.")

        for r in range(r_start, n_rand):
            alpha_zero = self._rand_vect_list[r]

            one_moment = [0.] * n_moments
            if new_rand_vect > 0:
                alpha = alpha_zero
                alpha_next = self.hamiltonian.matvec(alpha)
                if self.operator is None:
                    one_moment[0] = np.vdot(alpha_zero, alpha_zero)
                    one_moment[1] = np.vdot(alpha_zero, alpha_next)
                else:
                    one_moment[0] = self.operator(alpha_zero, alpha_zero)
                    one_moment[1] = self.operator(alpha_zero, alpha_next)

            if new_moments > 0:
                (alpha, alpha_next) = self._last_two_alphas[r]
                one_moment[0:self.num_moments] = self._moments_list[r]
            # Iteration over the moments
            # Two cases can occur, depicted in Eq. (28) and in Eq. (29),
            # respectively.
            # ----
            # In the first case, self.operator is None and we can use
            # Eqs. (34) and (35) to obtain the density of states, with
            # two moments ``one_moment`` for every new alpha.
            # ----
            # In the second case, the operator is not None and a matrix
            # multiplication should be used.
            if self.operator is None:
                for n in range(m_start//2, n_moments//2):
                    alpha_save = alpha_next
                    alpha_next = (2 * self.hamiltonian.matvec(alpha_next)
                                  - alpha)
                    alpha = alpha_save
                    # Following Eqs. (34) and (35)
                    one_moment[2*n] = (2 * np.vdot(alpha, alpha)
                                       - one_moment[0])
                    one_moment[2*n+1] = (2 * np.vdot(alpha_next, alpha)
                                         - one_moment[1])
                if n_moments % 2:
                    # odd moment
                    one_moment[n_moments - 1] = (
                        2 * np.vdot(alpha_next, alpha_next) - one_moment[0])
            else:
                for n in range(m_start, n_moments):
                    alpha_save = alpha_next
                    alpha_next = (2 * self.hamiltonian.matvec(alpha_next)
                                  - alpha)
                    alpha = alpha_save
                    one_moment[n] = self.operator(alpha_zero, alpha_next)

            self._last_two_alphas[r] = (alpha, alpha_next)
            self._moments_list[r] = one_moment[:]


# ### Auxiliary functions


def _rescale(hamiltonian, eps, v0, bounds):
    """Rescale a Hamiltonian and return a LinearOperator

    Parameters
    ----------
    hamiltonian : 2D array
        Hamiltonian of the system.
    eps : scalar
        Ensures that the bounds 'a' and 'b' are strict.
    v0 : random vector, or None
        Used as the initial residual vector for the algorithm that
        finds the lowest and highest eigenvalues.
    bounds : tuple, or None
        Boundaries of the spectrum. If not provided the maximum and
        minimum eigenvalues are calculated.
    """
    # Relative tolerance to which to calculate eigenvalues.  Because after
    # rescaling we will add eps / 2 to the spectral bounds, we don't need
    # to know the bounds more accurately than eps / 2.
    tol = eps / 2

    if bounds:
        lmin, lmax = bounds
    else:
        lmax = float(sla.eigsh(hamiltonian, k=1, which='LA',
                               return_eigenvectors=False, tol=tol, v0=v0))
        lmin = float(sla.eigsh(hamiltonian, k=1, which='SA',
                               return_eigenvectors=False, tol=tol, v0=v0))

    a = np.abs(lmax-lmin) / (2. - eps)
    b = (lmax+lmin) / 2.

    if lmax - lmin <= abs(lmax + lmin) * tol / 2:
        raise ValueError(
            'The Hamiltonian has a single eigenvalue, it is not possible to '
            'obtain a spectral density.')

    def rescaled(v):
        return (hamiltonian.dot(v) - b * v) / a

    rescaled_ham = sla.LinearOperator(shape=hamiltonian.shape, matvec=rescaled)

    return rescaled_ham, (a, b)


def _calc_fft_moments(moments, n_sampling):
    """This function takes the normalised moments and returns an array
    of points and an array of the evaluated function at those points.
    """
    moments = np.asarray(moments)
    # extra_shape handles the case where operators have vector outputs
    n_moments, *extra_shape = moments.shape
    moments_ext = np.zeros([n_sampling] + extra_shape)

    # Jackson kernel, as in Eq. (71), and kernel improved moments,
    # as in Eq. (81).
    m = np.arange(n_moments)
    kernel = ((n_moments - m + 1) * np.cos(np.pi * m / (n_moments + 1)) +
              np.sin(np.pi * m / (n_moments + 1)) /
              np.tan(np.pi / (n_moments + 1))) / (n_moments + 1)

    # special points at the abscissas of Chebyshev integration
    k = np.arange(0, n_sampling)
    xk_rescaled = np.cos(np.pi * (k + 0.5) / n_sampling)
    # prefactor in Eq. (80)
    gk = np.pi * np.sqrt(1 - xk_rescaled ** 2)

    # transposes handle the case where operators have vector outputs
    moments_ext[:n_moments] = np.transpose(moments.transpose() * kernel)
    # The function evaluated in these special data points is the FFT of
    # the moments as in Eq. (83).
    gammas = np.transpose(fft.dct(moments_ext.transpose(), type=3))

    # Element-wise division of moments_FFT over gk, as in Eq. (83).
    rho = np.transpose(np.divide(gammas.transpose(), gk))

    # Reverse energies and densities to set ascending order.
    return xk_rescaled[::-1], rho[::-1], gammas[::-1]



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['lll', 'cvp', 'voronoi']

import numpy as np
from itertools import product


def gs_coefficient(a, b):
    """Gram-Schmidt coefficient."""
    return np.dot(a, b) / np.linalg.norm(b)**2


def gs(mat):
    """Compute Gram-Schmidt decomposition on a matrix."""
    mat = np.copy(mat)
    for i in range(len(mat)):
        for j in range(i):
            mat[i] -= gs_coefficient(mat[i], mat[j]) * mat[j]
    return mat


def is_c_reduced(vecs, c):
    """Check if a basis is c-reduced."""
    vecs = gs(vecs)
    r = np.apply_along_axis(lambda x: np.linalg.norm(x)**2, 1, vecs)
    return np.all((r[: -1] / r[1:]) < c)


def lll(basis, c=1.34):
    """
    Calculate a reduced lattice basis using LLL algorithm.

    Reduce a basis of a lattice to an almost orthonormal form. For details see
    e.g. http://en.wikipedia.org/wiki/LLL-algorithm.

    Parameters
    ----------
    basis : 2d array-like of floats
        The lattice basis vectors to be reduced.
    c : float
        Reduction parameter for the algorithm. Must be larger than 1 1/3,
        since otherwise a solution is not guaranteed to exist.

    Returns
    -------
    reduced_basis : numpy array
        The basis vectors of the LLL-reduced basis.
    transformation : numpy integer array
        Coefficient matrix for tranforming from the reduced basis to the
        original one.
    """
    vecs_orig = np.asarray(basis, float)
    if vecs_orig.ndim != 2:
        raise ValueError('"basis" must be a 2d array-like object.')
    if vecs_orig.shape[0] > vecs_orig.shape[1]:
        raise ValueError('The number of basis vectors exceeds the '
                         'space dimensionality.')
    vecs = vecs_orig.copy()
    vecsstar = vecs_orig.copy()
    m = vecs.shape[0]
    u = np.identity(m)
    def ll_reduce(i):
        for j in reversed(range(i)):
            vecs[i] -= np.round(u[i, j]) * vecs[j]
            u[i] -= np.round(u[i, j]) * u[j]

    # Initialize values.
    for i in range(m):
        for j in range(i):
            u[i, j] = gs_coefficient(vecs[i], vecsstar[j])
            vecsstar[i] -= u[i, j] * vecsstar[j]
        ll_reduce(i)

    # Main part of LLL algorithm.
    i = 0
    while i < m-1:
        if (np.linalg.norm(vecsstar[i]) ** 2 <
           c * np.linalg.norm(vecsstar[i+1]) ** 2):
            i += 1
        else:
            vecsstar[i+1] += u[i+1, i] * vecsstar[i]

            u[i, i] = gs_coefficient(vecs[i], vecsstar[i+1])
            u[i, i+1] = u[i+1, i] = 1
            u[i+1, i+1] = 0
            vecsstar[i] -= u[i, i] * vecsstar[i+1]
            vecs[[i, i+1]] = vecs[[i+1, i]]
            vecsstar[[i, i+1]] = vecsstar[[i+1, i]]
            u[[i, i+1]] = u[[i+1, i]]
            for j in range(i+2, m):
                u[j, i] = gs_coefficient(vecs[j], vecsstar[i])
                u[j, i+1] = gs_coefficient(vecs[j], vecsstar[i+1])
            if abs(u[i+1, i]) > 0.5:
                ll_reduce(i+1)
            i = max(i-1, 0)
    # TODO: change to rcond=None once we depend on numpy >= 1.14.
    coefs = np.linalg.lstsq(vecs_orig.T, vecs.T, rcond=-1)[0]
    if not np.allclose(np.round(coefs), coefs, atol=1e-6):
        raise RuntimeError('LLL algorithm instability.')
    if not is_c_reduced(vecs, c):
        raise RuntimeError('LLL algorithm instability.')
    return vecs, np.array(np.round(coefs), int)


def cvp(vec, basis, n=1):
    """
    Solve the n-closest vector problem for a vector, given a basis.

    This algorithm performs poorly in general, so it should be supplied
    with LLL-reduced bases.

    Parameters
    ----------
    vec : 1d array-like of floats
        The lattice vectors closest to this vector are to be found.
    basis : 2d array-like of floats
        Sequence of basis vectors
    n : int
        Number of lattice vectors closest to the point that need to be found.

    Returns
    -------
    coords : numpy array
        An array with the coefficients of the lattice vectors closest to the
        requested point.

    Notes
    -----
    This function can also be used to solve the `n` shortest lattice vector
    problem if the `vec` is zero, and `n+1` points are requested
    (and the first output is ignored).
    """
    # Calculate coordinates of the starting point in this basis.
    basis = np.asarray(basis)
    if basis.ndim != 2:
        raise ValueError('`basis` must be a 2d array-like object.')
    vec = np.asarray(vec)
    # TODO: change to rcond=None once we depend on numpy >= 1.14.
    center_coords = np.linalg.lstsq(basis.T, vec, rcond=-1)[0]
    center_coords = np.array(np.round(center_coords), int)
    # Cutoff radius for n-th nearest neighbor.
    rad = 1
    nth_dist = np.inf
    while True:
        r = np.round(rad * np.linalg.cond(basis)) + 1
        points = np.mgrid[tuple(slice(i - r, i + r) for i in center_coords)]
        points = points.reshape(basis.shape[0], -1).T
        if len(points) < n:
            rad += 1
            continue
        point_coords = np.dot(points, basis)
        point_coords -= vec.T
        distances = np.sqrt(np.sum(point_coords**2, 1))
        order = np.argsort(distances)
        new_nth_dist = distances[order[n - 1]]
        if new_nth_dist < nth_dist:
            nth_dist = new_nth_dist
            rad += 1
        else:
            return np.array(points[order[:n]], int)


def voronoi(basis):
    """
    Return an array of lattice vectors forming its voronoi cell.

    Parameters
    ----------
    basis : 2d array-like of floats
        Basis vectors for which the Voronoi neighbors have to be found.

    Returns
    -------
    voronoi_neighbors : numpy array of ints
        All the lattice vectors that may potentially neighbor the origin.

    Notes
    -----
    This algorithm does not calculate the minimal Voronoi cell of the lattice
    and can be optimized. Its main aim is flood-fill, however, and better
    safe than sorry.
    """
    basis = np.asarray(basis)
    if basis.ndim != 2:
        raise ValueError('`basis` must be a 2d array-like object.')
    displacements = list(product(*(len(basis) * [[0, .5]])))[1:]
    vertices = np.array([cvp(np.dot(vec, basis), basis)[0] for vec in
                         displacements])
    vertices = np.array(np.round((vertices - displacements) * 2), int)
    for i in range(len(vertices)):
        if not np.any(vertices[i]):
            vertices[i] += 2 * np.array(displacements[i])
    vertices = np.concatenate([vertices, -vertices])
    return vertices



# Copyright 2011-2013 Kwant authors.
#
# This file is part of Kwant.  It is subject to the license terms in the file
# LICENSE.rst found in the top-level directory of this distribution and at
# http://kwant-project.org/license.  A list of Kwant authors can be found in
# the file AUTHORS.rst at the top-level directory of this distribution and at
# http://kwant-project.org/authors.

__all__ = ['schur', 'convert_r2c_schur', 'order_schur', 'evecs_from_schur',
           'gen_schur', 'order_gen_schur', 'convert_r2c_gen_schur',
           'evecs_from_gen_schur']

from math import sqrt
import numpy as np
from . import lapack


def schur(a, calc_q=True, calc_ev=True, overwrite_a=False):
    """Compute the Schur form of a square matrix a.

    The Schur form is a decomposition of the form a = q * t * q^dagger, where q
    is a unitary matrix and t a upper triagonal matrix when computing the Schur
    form of a complex matrix, and a quasi-upper triagonal matrix with only 1x1
    and 2x2 blocks on the diagonal when computing the Schur form of a real
    matrix (In the latter case, the 1x1 blocks correspond to real eigenvalues,
    the 2x2 blocks to conjugate pairs of complex eigenvalues).

    The Schur form is closely related to the eigenvalue problem (the entries of
    the diagonal of the complex Schur form are the eigenvalues of the matrix),
    and the routine can optionally also return the eigenvalues.

    Parameters
    ----------
    a : array, shape (M, M)
        Matrix for which to compute the Schur form.
    calc_q : boolean
        Whether to compute the unitary/orthogonal matrix `q`.
    calc_ev : boolean
        Whether to return the eigenvalues as a separate array.
    overwrite_a : boolean
        Whether to overwrite data in `a` (may increase performance).

    Returns
    -------
    t : array, shape (M, M)
        Schur form of the original matrix (complex or real, depending on the
        input matrix).

    (if calc_q == True)
    q : array, shape (M, M)
        Unitary transformation matrix.

    (if calc_ev == True)
    ev: array, shape (M,)
        Array of eigenvalues of the matrix `a`. Can be complex even if a is
        real. In the latter case, the complex eigenvalues come in conjugated
        pairs with the eigenvalue with positive imaginary part coming
        first.

    Raises
    ------
    LinAlgError
        If the underlying QR iteration fails to converge.
    """
    a = lapack.prepare_for_lapack(overwrite_a, a)
    return lapack.gees(a, calc_q, calc_ev)


def convert_r2c_schur(t, q):
    """Convert a real Schur form (with possibly 2x2 blocks on the diagonal)
    into a complex Schur form that is completely triangular.

    This function is equivalent to the scipy.linalg.rsf2csf pendant (though the
    implementation is different), but there is additionally the guarantee that
    in the case of a 2x2 block at rows and columns i and i+1, t[i, i] will
    contain the eigenvalue with the positive part, and t[i+1, i+1] the one with
    the negative part.  This ensures that the list of eigenvalues (more
    precisely, their order) returned originally from schur() is still valid for
    the newly formed complex Schur form.

    Parameters
    ----------
    t : array, shape (M, M)
        Real Schur form of the original matrix
    q : array, shape (M, M)
        Schur transformation matrix

    Returns
    -------
    t : array, shape (M, M)
        Complex Schur form of the original matrix
    q : array, shape (M, M)
        Schur transformation matrix corresponding to the complex form
    """

    # First find the positions of 2x2-blocks
    blockpos = np.diagonal(t, -1).nonzero()[0]

    # Check if there are actually any 2x2-blocks
    if not blockpos.size:
        return (t, q)
    else:
        t2 = t.astype(np.common_type(t, np.array([], np.complex64)))
        q2 = q.astype(np.common_type(q, np.array([], np.complex64)))

    for i in blockpos:
        # Bringing a 2x2 block to complex triangular form is relatively simple:
        # the 2x2 blocks are guaranteed to be of the form [[a, b], [c, a]],
        # where b*c < 0. The eigenvalues of this matrix are a +/- i sqrt(-b*c),
        # the corresponding eigenvectors are [ +/- sqrt(-b*c), c].  The Schur
        # form can be achieved by a unitary 2x2 matrix with one of the
        # eigenvectors in the first column, and the second column an orthogonal
        # vector.

        a = t[i, i]
        b = t[i, i+1]
        c = t[i+1, i]

        x = 1j * sqrt(-b * c)
        y = c
        norm = sqrt(-b * c + c * c)

        U = np.array([[x / norm, -y / norm], [y / norm, -x / norm]])

        t2[i, i] = a + x
        t2[i+1, i] = 0
        t2[i, i+1] = -b - c
        t2[i+1, i+1] = a - x

        t2[:i, i:i+2] = np.dot(t2[:i, i:i+2], U)
        t2[i:i+2, i+2:] = np.dot(np.conj(U.T), t2[i:i+2, i+2:])

        q2[:, i:i+2] = np.dot(q2[:, i:i+2], U)

    return t2, q2


def order_schur(select, t, q, calc_ev=True, overwrite_tq=False):
    """Reorder the Schur form, selecting a cluster of eigenvalues.

    This function reorders the generalized Schur form such that the cluster of
    eigenvalues determined by select appears in the leading diagonal block of
    the Schur form (this is useful, as the Schur vectors corresponding to the
    leading diagonal block form an orthogonal basis for the subspace of
    eigenvectors).

    If a real Schur form is reordered, it is converted to complex form
    (eliminating the 2x2 blocks on the diagonal) if in a complex conjugated
    pair of eigenvalues only one eigenvalue is chosen.  In this case, the real
    Schur from cannot be reordered in real form without splitting a 2x2 block
    on the diagonal, hence switching to complex form is mandatory.

    Parameters
    ----------
    t : array, shape (M, M)
        Schur form
    q : array, shape (M, M)
        Unitary/orthogonal transformation matrices.
    calc_ev : boolean, optional
        Whether to return the reordered generalized eigenvalues of as two
        separate arrays. Default: True
    overwrite_tq : boolean, optional
        Whether to overwrite data in `t` and `q` (may increase performance)
        Default: False

    Returns
    -------
    t : array, shape (M, M)
        Reordered Schur form. If the original Schur form is real, and the
        desired reordering separates complex conjugated pairs of generalized
        eigenvalues, the resulting Schur form will be complex.
    q : array, shape (M, M)
        Unitary/orthogonal transformation matrix. Only computed if q is
        provided (not None) as input. If the Schur form is converted from real
        to complex, the transformation matrix is also converted from real
        orthogonal to complex unitary
    alpha : array, shape (M)
    beta : array, shape (M)
        Reordered eigenvalues. If the reordered Schur form is real, complex
        conjugated pairs of eigenvalues are ordered such that the eigenvalue
        with the positive imaginary part comes first.  Only computed if
        ``calc_ev == True``
    """

    t, q = lapack.prepare_for_lapack(overwrite_tq, t, q)

    # Figure out if select is a function or array.
    isfun = isarray = True
    try:
        select(0)
    except:
        isfun = False
    try:
        select[0]
    except:
        isarray = False

    if not (isarray or isfun):
        raise ValueError("select must be either a function or an array")
    elif isarray:
        select = np.array(select, dtype=lapack.logical_dtype, order='F')
    else:
        select = np.array(np.vectorize(select)(np.arange(t.shape[0])),
                          dtype=lapack.logical_dtype, order='F')

    # Now check if the reordering can actually be done as desired,
    # if we have a real Schur form (i.e. if the 2x2 blocks would be
    # separated). If this is the case, convert to complex Schur form first.
    for i in np.diagonal(t, -1).nonzero()[0]:
        if bool(select[i]) != bool(select[i+1]):
            t, q = convert_r2c_schur(t, q)
            return order_schur(select, t, q, calc_ev, True)

    return lapack.trsen(select, t, q, calc_ev)


def evecs_from_schur(t, q, select=None, left=False, right=True,
                     overwrite_tq=False):
    """Compute eigenvectors from Schur form.

    This function computes either all or selected eigenvectors for the matrix
    that is represented by the Schur form t and the unitary matrix q, (not the
    eigenvectors of t, but of q*t*q^dagger).

    Parameters
    ----------
    t : array, shape (M, M)
        Schur form
    q : array, shape (M, M)
        Unitary/orthogonal transformation matrix.
    select : boolean function or array, optional
        The value of ``select(i)`` or ``select[i]`` is used to decide whether
        the eigenvector corresponding to the i-th eigenvalue should be
        computed or not. If select is not provided (None), all eigenvectors
        are computed. Default: None
    left : boolean, optional
        Whether to compute left eigenvectors. Default: False
    right : boolean, optional
        Whether to compute right eigenvectors. Default: True
    overwrite_tq : boolean, optional
        Whether to overwrite data in `t` and `q` (may increase performance)
        Default: False

    Returns
    -------
    vl : array, shape(M, N)
        Left eigenvectors. N is the number of eigenvectors selected b
        `select`, or equal to M if select is not provided. The eigenvectors
        may be complex, even if `t` and `q` are real. Only computed if
        ``left == True``.
    vr : array, shape(M, N)
        Right eigenvectors. N is the number of eigenvectors selected by
        `select`, or equal to M if select is not provided. The eigenvectors
        may be complex, even if `t` and `q` are real. Only computed if
        ``right == True``.
    """

    t, q = lapack.prepare_for_lapack(overwrite_tq, t, q)

    # check if select is a function or an array
    if select is not None:
        isfun = isarray = True
        try:
            select(0)
        except:
            isfun = False

        try:
            select[0]
        except:
            isarray = False

        if not (isarray or isfun):
            raise ValueError("select must be either a function, "
                             "an array or None")
        elif isarray:
            selectarr = np.array(select, dtype=lapack.logical_dtype,
                                 order='F')
        else:
            selectarr = np.array(np.vectorize(select)(np.arange(t.shape[0])),
                                 dtype=lapack.logical_dtype, order='F')
    else:
        selectarr = None

    return lapack.trevc(t, q, selectarr, left, right)


def gen_schur(a, b, calc_q=True, calc_z=True, calc_ev=True,
              overwrite_ab=False):
    """Compute the generalized Schur form of a matrix pencil (a, b).

    The generalized Schur form is a decomposition of the form a = q * s *
    z^dagger and b = q * t * z^dagger, where q and z are unitary matrices
    (orthogonal for real input), t is an upper triagonal matrix with
    non-negative real diagonal, and s is a upper triangular matrix for complex
    matrices, and a quasi-upper triangular matrix with only 1x1 and 2x2 blocks
    on the diagonal for real matrices.  (In the latter case, the 1x1 blocks
    correspond to real generalized eigenvalues, the 2x2 blocks to conjugate
    pairs of complex generalized eigenvalues).

    The generalized Schur form is closely related to the generalized eigenvalue
    problem (the entries of the diagonal of the complex Schur form are the
    eigenvalues of the matrix, for example), and the routine can optionally
    also return the generalized eigenvalues in the form (alpha, beta), such
    that alpha/beta is a generalized eigenvalue of the pencil (a, b) (see also
    gen_eig()).

    Parameters
    ----------
    a : array, shape (M, M)
    b : array, shape (M, M)
        Matrix pencil for which to compute the generalized Schur form
    calc_q : boolean, optional
    calc_z : boolean, optional
        Whether to compute the unitary/orthogonal matrices `q` and `z`.
        Default: True
    calc_ev : boolean, optional
        Whether to return the generalized eigenvalues as two separate
        arrays. Default: True
    overwrite_ab : boolean, optional
        Whether to overwrite data in `a` and `b` (may increase performance)
        Default: False

    Returns
    -------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Generalized Schur form of the original matrix pencil (`a`,`b`)
        (complex or real, depending on the input matrices)
    q : array, shape (M, M)
    z : array, shape (M, M)
        Unitary/orthogonal transformation matrices. Only computed if
        ``calc_q == True`` or ``calc_z == True``, respectively.
    alpha : array, shape (M)
    beta : array, shape (M)
        Generalized eigenvalues of the matrix pencil (`a`, `b`) given
        as numerator (`alpha`) and denominator (`beta`), such that the
        generalized eigenvalues are given as ``alpha/beta``. alpha can
        be complex even if a is real. In the latter case, complex
        eigenvalues come in conjugated pairs with the eigenvalue with
        positive imaginary part coming first. Only computed if
        ``calc_ev == True``.

    Raises
    ------
    LinAlError
        If the underlying QZ iteration fails to converge.
    """
    a, b = lapack.prepare_for_lapack(overwrite_ab, a, b)
    return lapack.gges(a, b, calc_q, calc_z, calc_ev)


def order_gen_schur(select, s, t, q=None, z=None, calc_ev=True,
                    overwrite_stqz=False):
    """Reorder the generalized Schur form.

    This function reorders the generalized Schur form such that the cluster of
    eigenvalues determined by select appears in the leading diagonal blocks of
    the Schur form (this is useful, as the Schur vectors corresponding to the
    leading diagonal blocks form an orthogonal basis for the subspace of
    eigenvectors).

    If a real generalized Schur form is reordered, it is converted to complex
    form (eliminating the 2x2 blocks on the diagonal) if in a complex
    conjugated pair of eigenvalues only one eigenvalue is chosen.  In this
    case, the real Schur from cannot be reordered in real form without
    splitting a 2x2 block on the diagonal, hence switching to complex form is
    mandatory.

    Parameters
    ----------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Matrices describing the generalized Schur form.
    q : array, shape (M, M), optional
    z : array, shape (M, M), optional
        Unitary/orthogonal transformation matrices. Default: None.
    calc_ev : boolean, optional
        Whether to return the reordered generalized eigenvalues of as two
        separate arrays. Default: True.
    overwrite_stqz : boolean, optional
        Whether to overwrite data in `s`, `t`, `q`, and `z` (may
        increase performance) Default: False.

    Returns
    -------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Reordered general Schur form. If the original Schur form is real, and
        the desired reordering separates complex conjugated pairs of
        generalized eigenvalues, the resulting Schur form will be complex.
    q : array, shape (M, M)
    z : array, shape (M, M)
        Unitary/orthogonal transformation matrices. Only computed if
        `q` and `z` are provided (not None) on entry, respectively. If
        the generalized Schur form is converted from real to complex,
        the transformation matrices are also converted from real
        orthogonal to complex unitary
    alpha : array, shape (M)
    beta : array, shape (M)
        Reordered generalized eigenvalues. If the reordered Schur form is real,
        complex conjugated pairs of eigenvalues are ordered such that the
        eigenvalue with the positive imaginary part comes first.  Only computed
        if ``calc_ev == True``.

    Raises
    ------
    LinAlError
        If the problem is too ill-conditioned.
    """
    s, t, q, z = lapack.prepare_for_lapack(overwrite_stqz, s, t, q, z)


    # Figure out if select is a function or array.
    isfun = isarray = True
    try:
        select(0)
    except:
        isfun = False
    try:
        select[0]
    except:
        isarray = False

    if not (isarray or isfun):
        raise ValueError("select must be either a function or an array")
    elif isarray:
        select = np.array(select, dtype=lapack.logical_dtype, order='F')
    else:
        select = np.array(np.vectorize(select)(np.arange(t.shape[0])),
                          dtype=lapack.logical_dtype, order='F')

    # Now check if the reordering can actually be done as desired, if we have a
    # real Schur form (i.e. if the 2x2 blocks would be separated). If this is
    # the case, convert to complex Schur form first.
    for i in np.diagonal(s, -1).nonzero()[0]:
        if bool(select[i]) != bool(select[i+1]):
            # Convert to complex Schur form
            if q is not None and z is not None:
                s, t, q, z = convert_r2c_gen_schur(s, t, q, z)
            elif q is not None:
                s, t, q = convert_r2c_gen_schur(s, t, q=q, z=None)
            elif z is not None:
                s, t, z = convert_r2c_gen_schur(s, t, q=None, z=z)
            else:
                s, t = convert_r2c_gen_schur(s, t)

            return order_gen_schur(select, s, t, q, z, calc_ev, True)

    return lapack.tgsen(select, s, t, q, z, calc_ev)


def convert_r2c_gen_schur(s, t, q=None, z=None):
    """Convert a real generallzed Schur form (with possibly 2x2 blocks on the
    diagonal) into a complex Schur form that is completely triangular.  If the
    input is already completely triagonal (real or complex), the input is
    returned unchanged.

    This function guarantees that in the case of a 2x2 block at rows and
    columns i and i+1, the converted, complex Schur form will contain the
    generalized eigenvalue with the positive imaginary part in s[i,i] and
    t[i,i], and the one with the negative imaginary part in s[i+1,i+1] and
    t[i+1,i+1].  This ensures that the list of eigenvalues (more precisely,
    their order) returned originally from gen_schur() is still valid for the
    newly formed complex Schur form.

    Parameters
    ----------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Real generalized Schur form of the original matrix
    q : array, shape (M, M), optional
    z : array, shape (M, M), optional
        Schur transformation matrix. Default: None

    Returns
    -------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Complex generalized Schur form of the original matrix,
        completely triagonal
    q : array, shape (M, M)
    z : array, shape (M, M)
        Schur transformation matrices corresponding to the complex
        form. `q` or `z` are only computed if they are provided (not
        None) on input.

    Raises
    ------
    LinAlgError
        If it fails to convert a 2x2 block into complex form (unlikely).
    """

    s, t, q, z = lapack.prepare_for_lapack(True, s, t, q, z)
    # Note: overwrite=True does not mean much here, the arrays are all copied

    if (s.ndim != 2 or t.ndim != 2 or
        (q is not None and q.ndim != 2) or
        (z is not None and z.ndim != 2)):
        raise ValueError("Expect matrices as input")

    if ((s.shape[0] != s.shape[1] or t.shape[0] != t.shape[1] or
         s.shape[0] != t.shape[0]) or
        (q is not None and (q.shape[0] != q.shape[1] or
                            s.shape[0] != q.shape[0])) or
        (z is not None and (z.shape[0] != z.shape[1] or
                            s.shape[0] != z.shape[0]))):
        raise ValueError("Invalid Schur decomposition as input")

    # First, find the positions of 2x2-blocks.
    blockpos = np.diagonal(s, -1).nonzero()[0]

    # Check if there are actually any 2x2-blocks.
    if not blockpos.size:
        s2 = s
        t2 = t
        q2 = q
        z2 = z
    else:
        s2 = s.astype(np.common_type(s, np.array([], np.complex64)))
        t2 = t.astype(np.common_type(t, np.array([], np.complex64)))
        if q is not None:
            q2 = q.astype(np.common_type(q, np.array([], np.complex64)))
        if z is not None:
            z2 = z.astype(np.common_type(z, np.array([], np.complex64)))

    for i in blockpos:
        # In the following, we use gen_schur on individual 2x2 blocks (that are
        # promoted to complex form) to compute the complex generalized Schur
        # form. If necessary, order_gen_schur is used to ensure the desired
        # order of eigenvalues.

        sb, tb, qb, zb, alphab, betab = gen_schur(s2[i:i+2, i:i+2],
                                                  t2[i:i+2, i:i+2])

        # Ensure order of eigenvalues. (betab is positive)
        if alphab[0].imag < alphab[1].imag:
            sb, tb, qb, zb, alphab, betab = order_gen_schur([False, True],
                                                            sb, tb, qb, zb)

        s2[i:i+2, i:i+2] = sb
        t2[i:i+2, i:i+2] = tb

        s2[:i, i:i+2] = np.dot(s2[:i, i:i+2], zb)
        s2[i:i+2, i+2:] = np.dot(qb.T.conj(), s2[i:i+2, i+2:])
        t2[:i, i:i+2] = np.dot(t2[:i, i:i+2], zb)
        t2[i:i+2, i+2:] = np.dot(qb.T.conj(), t2[i:i+2, i+2:])

        if q is not None:
            q2[:, i:i+2] = np.dot(q[:, i:i+2], qb)
        if z is not None:
            z2[:, i:i+2] = np.dot(z[:, i:i+2], zb)

    if q is not None and z is not None:
        return s2, t2, q2, z2
    elif q is not None:
        return s2, t2, q2
    elif z is not None:
        return s2, t2, z2
    else:
        return s2, t2


def evecs_from_gen_schur(s, t, q=None, z=None, select=None,
                         left=False, right=True, overwrite_qz=False):
    """Compute eigenvectors from Schur form.

    This function computes either all or selected eigenvectors for the matrix
    that is represented by the generalized Schur form (s, t) and the unitary
    matrices q and z, (not the generalized eigenvectors of (s,t), but of
    (q*s*z^dagger, q*t*z^dagger)).

    Parameters
    ----------
    s : array, shape (M, M)
    t : array, shape (M, M)
        Generalized Schur form.
    q : array, shape (M, M), optional
    z : array, shape (M, M), optional
        Unitary/orthogonal transformation matrices. If the left eigenvectors
        are to be computed, `q` must be provided, if the right eigenvectors are
        to be computed, `z` must be provided.
    select : boolean function or array, optional
        The value of ``select(i)`` or ``select[i]`` is used to decide
        whether the eigenvector corresponding to the i-th eigenvalue
        should be computed or not. If select is not provided, all
        eigenvectors are computed. Default: None.
    left : boolean, optional
        Whether to compute left eigenvectors. Default: False.
    right : boolean, optional
        Whether to compute right eigenvectors. Default: True.
    overwrite_qz : boolean, optional
        Whether to overwrite data in `q` and `z` (may increase performance).
        Note that s and t remain always unchanged Default: False.

    Returns
    -------
    (if left == True)
    vl : array, shape(M, N)
        Left generalized eigenvectors. N is the number of eigenvectors
        selected by select, or equal to M if select is not
        provided. The eigenvectors may be complex, even if `s`, `t`,
        `q` and `z` are real.

    (if right == True)
    vr : array, shape(M, N)
        Right generalized eigenvectors. N is the number of
        eigenvectors selected by select, or equal to M if select is
        not provided. The eigenvectors may be complex, even if `s`,
        `t`, `q` and `z` are real.

    """

    s, t, q, z = lapack.prepare_for_lapack(overwrite_qz, s, t, q, z)

    if left and q is None:
        raise ValueError("Matrix q must be provided for left eigenvectors")

    if right and z is None:
        raise ValueError("Matrix z must be provided for right eigenvectors")

    # Check if select is a function or an array.
    if select is not None:
        isfun = isarray = True
        try:
            select(0)
        except:
            isfun = False

        try:
            select[0]
        except:
            isarray = False

        if not (isarray or isfun):
            raise ValueError("select must be either a function, "
                             "an array or None")
        elif isarray:
            selectarr = np.array(select, dtype=lapack.logical_dtype,
                                 order='F')
        else:
            selectarr = np.array(np.vectorize(select)(np.arange(t.shape[0])),
                                 dtype=lapack.logical_dtype, order='F')
    else:
        selectarr = None

    return lapack.tgevc(s, t, q, z, selectarr, left, right)
